# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import pandas as pd

from openstef.data_classes.model_specifications import ModelSpecificationDataClass
from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.model.confidence_interval_applicator import ConfidenceIntervalApplicator
from openstef.model.regressors.regressor import OpenstfRegressor
from openstef.model_selection.model_selection import backtest_split_default
from openstef.pipeline import train_model
from openstef.postprocessing.postprocessing import (
    add_prediction_job_properties_to_forecast,
)

DEFAULT_TRAIN_HORIZONS: list[float] = [0.25, 24.0]
DEFAULT_EARLY_STOPPING_ROUNDS: int = 10


def train_model_and_forecast_back_test(
    pj: PredictionJobDataClass,
    modelspecs: ModelSpecificationDataClass,
    input_data: pd.DataFrame,
    training_horizons: list[float] = None,
    n_folds: int = 1,
) -> tuple[
    pd.DataFrame,
    list[OpenstfRegressor],
    list[pd.DataFrame],
    list[pd.DataFrame],
    list[pd.DataFrame],
]:
    """Pipeline for a back test.

    When number of folds is larger than 1: apply pipeline for a back test when forecasting
    the entire input range.

        - Makes use of kfold cross validation in order to split data multiple times.
        - Results of all the testsets are added together to obtain the forecast for the whole input range.
        - Obtaining the days for each fold can be done either randomly or not
        **DO NOT USE THIS PIPELINE FOR OPERATIONAL FORECASTS**

    Args:
        pj: Prediction job.
        modelspecs: Dataclass containing model specifications
        input_data: Input data
        training_horizons: horizons to train on in hours.
            These horizons are also used to make predictions (one for every horizon)
        n_folds: number of folds to apply (if 1, no cross validation will be applied)

    Returns:
        - Forecast (pandas.DataFrame)
        - Fitted models (list[OpenStfRegressor])
        - Train data sets (list[pd.DataFrame])
        - Validation data sets (list[pd.DataFrame])
        - Test data sets (list[pd.DataFrame])

    """
    if pj.backtest_split_func is None:
        backtest_split_func = backtest_split_default
        backtest_split_args = {"stratification_min_max": pj["model"] != "proloaf"}
    else:
        backtest_split_func, backtest_split_args = pj.backtest_split_func.load(
            required_arguments=["data", "n_folds"]
        )

    data_with_features = train_model.train_pipeline_step_compute_features(
        input_data=input_data, pj=pj, model_specs=modelspecs, horizons=training_horizons
    )

    # The use of zip allows to take advantage of the lazy estimation mechanisms of Python, especially if the
    # backtest_split_func returns a generator. This can avoid unwanted multiple data copies.
    # 1. First we retrieve a generator (use of () comprehensive) on (model, forecast, train, val, test)
    # 2. Then we unzip the result into generators separated by result type (models, forecasts, trains, vals, tests)
    (
        models_folds,
        forecast_folds,
        train_data_folds,
        validation_data_folds,
        test_data_folds,
    ) = zip(
        *(
            train_model_and_forecast_test_core(
                pj, modelspecs, train_data, validation_data, test_data
            )
            + (train_data, validation_data, test_data)
            for train_data, validation_data, test_data in backtest_split_func(
                data_with_features, n_folds, **backtest_split_args
            )
        )
    )

    return (
        pd.concat(forecast_folds, axis=0).sort_index(),
        list(models_folds),
        list(train_data_folds),
        list(validation_data_folds),
        list(test_data_folds),
    )


def train_model_and_forecast_test_core(
    pj: PredictionJobDataClass,
    modelspecs: ModelSpecificationDataClass,
    train_data: pd.DataFrame,
    validation_data: pd.DataFrame,
    test_data: pd.DataFrame,
) -> tuple[OpenstfRegressor, pd.DataFrame]:
    """Trains the model and forecast on the test set.

    Args:
        pj: Prediction job.
        modelspecs: Dataclass containing model specifications
        train_data: Train data with computed features
        validation_data: Validation data with computed features
        test_data: Test data with computed features

    Returns:
        - The trained model
        - The forecast on the test set.

    """
    model = train_model.train_pipeline_step_train_model(
        pj, modelspecs, train_data, validation_data
    )

    # Predict
    model_forecast = model.predict(test_data.iloc[:, 1:-1])
    forecast = pd.DataFrame(index=test_data.index, data={"forecast": model_forecast})

    # Define tAhead to something meaningfull in the context of a backtest
    forecast["tAhead"] = test_data.iloc[:, -1]

    # Add confidence
    forecast = ConfidenceIntervalApplicator(
        model, test_data.iloc[:, 1:-1]
    ).add_confidence_interval(forecast, pj)

    # Prepare for output
    forecast = add_prediction_job_properties_to_forecast(
        pj, forecast, algorithm_type="backtest"
    )

    # Add column with realised load and horizon information
    forecast["realised"] = test_data.iloc[:, 0]
    forecast["horizon"] = test_data.iloc[:, -1]

    return model, forecast



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# This file contains the Objective loss functions for quantile regression from:
# https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7
#
# SPDX-FileCopyrightText: 2017 Nikolay Lysenko
#
# SPDX-License-Identifier: MIT
"""This module contains all metrics to assess forecast quality."""
from typing import Callable

import numpy as np
import pandas as pd
import xgboost


def get_eval_metric_function(metric_name: str) -> Callable:
    """Gets a metric if it is available.

    Args:
        metric_name: Name of the metric.

    Returns:
        Function to calculate the metric.

    """
    evaluation_function = {
        "rmse": rmse,
        "bias": bias,
        "nsme": nsme,
        "mae": mae,
        "r_mae": r_mae,
        "frac_in_stdev": frac_in_stdev,
        "r_mae_highest": r_mae_highest,
        "r_mne_highest": r_mne_highest,
        "r_mpe_highest": r_mpe_highest,
        "r_mae_lowest": r_mae_lowest,
        "skill_score": skill_score,
        "skill_score_positive_peaks": skill_score_positive_peaks,
        "franks_skill_score": franks_skill_score,
        "franks_skill_score_peaks": franks_skill_score_peaks,
    }.get(metric_name, None)

    if evaluation_function is None:
        raise KeyError(f"Unknown evaluation metric function {metric_name}")

    return evaluation_function


def rmse(realised: pd.Series, forecast: pd.Series) -> float:
    """Function that calculates the Root Mean Square Error based on the true and prediciton.

    Args:
        realised: Realised load.
        forecast: Forecasted load.

    Returns:
        Root Mean Square Error

    """
    return np.sqrt(((realised - forecast) ** 2).mean())


def bias(realised: pd.Series, forecast: pd.Series) -> float:
    """Function that calculates the absolute bias in % based on the true and prediciton.

    Args:
        realised: Realised load.
        forecast: Forecasted load.

    Returns:
        Bias

    """
    return np.mean(forecast - realised)


def nsme(realised: pd.Series, forecast: pd.Series) -> float:
    """Function that calculates the Nash-sutcliffe model efficiency based on the true and prediciton.

    Args:
        realised: Realised load.
        forecast: Forecasted load.

    Returns:
        Nash-sutcliffe model efficiency

    """
    try:
        return 1 - sum((forecast - realised) ** 2) / sum(
            (realised - np.mean(realised)) ** 2
        )
    except ZeroDivisionError:  # means the error is 0
        return 1


def mae(realised: pd.Series, forecast: pd.Series) -> float:
    """Function that calculates the mean absolute error based on the true and prediction."""
    return np.mean(np.abs(forecast - realised))


def r_mae(realised: pd.Series, forecast: pd.Series) -> float:
    """Function that calculates the relative mean absolute error based on the true and prediction.

    The range is based on the load range of the previous two weeks

    """
    # Determine load range on entire dataset
    range_ = (
        realised.max() - realised.min()
        if (realised.max() - realised.min()) != 0
        else np.nan
    )

    return mae(realised, forecast) / range_


def frac_in_stdev(realised: pd.Series, forecast: pd.Series, stdev: pd.Series) -> float:
    """Function that calculates the amount of measurements that are within one stdev of our predictions."""
    outside_stdev = forecast[(forecast - realised).abs() > stdev]
    return round((1 - (len(outside_stdev) / len(forecast))), 2)


def r_mae_highest(
    realised: pd.Series, forecast: pd.Series, percentile: float = 0.95
) -> float:
    """Function that calculates the relative mean absolute error for the 5 percent highest realised values.

    The range is based on the load range of the previous two weeks.

    """
    # Check if length of both arrays is equal
    if len(np.array(realised)) != len(np.array(forecast)):
        raise ValueError(
            "Error metric can only be calculated for arrays of equal length!"
        )

    # Determine load range on entire dataset
    range_ = (
        realised.max() - realised.min()
        if (realised.max() - realised.min()) != 0
        else np.nan
    )

    # Get highest percentile of values
    highest_values = realised > np.percentile(realised, percentile)

    # Calculate mae
    r_mae_highest = mae(realised[highest_values], forecast[highest_values]) / range_

    return r_mae_highest


def r_mne_highest(realised: pd.Series, forecast: pd.Series) -> float:
    """Function that calculates the relative mean negative error for the 5 percent highest realised values.

    The range is based on the load range of the previous two weeks, this measure quantifies how much we underestimate
    peaks.

    """
    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast], axis=1)

    # Determine load range on entire dataset
    range_ = (
        combined["load"].max() - combined["load"].min()
        if (combined["load"].max() - combined["load"].min()) != 0
        else np.nan
    )

    # Select 5 percent highest realised load values
    combined["highest"] = combined["load"][
        combined["load"] > combined["load"].quantile(0.95)
    ]
    combined = combined[np.invert(np.isnan(combined["highest"]))]

    # Calculate rMNE for the selected points
    diff = combined[forecast.name] - combined["load"]

    if len(diff[diff < 0]) < 2:
        return 0.0

    r_mne_highest = np.mean(diff[diff < 0]) / range_

    if np.isnan(r_mne_highest):
        return 99999.0

    return r_mne_highest


def r_mpe_highest(realised: pd.Series, forecast: pd.Series) -> float:
    """Function that calculates the relative mean positive error for the 5 percent highest realised values.

    The range is based on the load range of the previous two weeks, this measure quantifies how much we overestimate
    peaks.

    """
    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast], axis=1)

    # Determine load range on entire dataset
    range_ = (
        combined["load"].max() - combined["load"].min()
        if (combined["load"].max() - combined["load"].min()) != 0
        else np.nan
    )

    # Select 5 percent highest realised load values
    combined["highest"] = combined["load"][
        combined["load"] > combined["load"].quantile(0.95)
    ]
    combined = combined[np.invert(np.isnan(combined["highest"]))]

    # Calculate rMPE for the selected points

    diff = combined[forecast.name] - combined["load"]

    if len(diff[diff > 0]) < 2:
        return 0.0

    r_mpe_highest = np.mean(diff[diff > 0]) / range_

    if np.isnan(r_mpe_highest):
        return 99999.0
    return r_mpe_highest


def r_mae_lowest(
    realised: pd.Series, forecast: pd.Series, quantile: float = 0.05
) -> float:
    """Function that calculates the relative mean absolute error for the 5 percent lowest realised values.

    The range is based on the load range of the previous two weeks.

    """
    # Determine load range on entire dataset
    range_ = (
        realised.max() - realised.min()
        if (realised.max() - realised.min()) != 0
        else np.nan
    )

    # Get lowest percentile of values
    lowest_values = realised < np.quantile(realised, quantile)
    # Calculate mae
    r_mae_lowest = mae(realised[lowest_values], forecast[lowest_values]) / range_

    return r_mae_lowest


def skill_score(realised: pd.Series, forecast: pd.Series, mean: pd.Series) -> float:
    """Function that calculates the skill score.

    Thise indicates model performance relative to a reference, in this case the mean of the realised values. The range
    is based on the load range of the previous two weeks.

    """
    combined = pd.concat([realised, forecast], axis=1)
    combined["mean"] = mean

    skill_score = 1 - (mae(realised, forecast) / mae(realised, combined["mean"]))

    if np.isnan(skill_score):
        return 0

    return skill_score


def skill_score_positive_peaks(
    realised: pd.Series, forecast: pd.Series, mean: pd.Series
) -> float:
    """Calculates skill score on positive peaks."""
    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast], axis=1)

    # Select 5 percent highest realised load values
    combined["highest"] = combined["load"][
        combined["load"] > combined["load"].quantile(0.95)
    ]
    combined = combined[np.invert(np.isnan(combined["highest"]))]

    # Calculate rMAE for the selected points
    skill_score_highest = skill_score(combined["load"], combined[forecast.name], mean)

    if np.isnan(skill_score_highest):
        return 0

    return skill_score_highest


def franks_skill_score(
    realised: pd.Series, forecast: pd.Series, basecase: pd.Series, range_: float = 1.0
) -> float:
    """Calculate Franks skill score."""
    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast], axis=1)
    if range_ == 1.0:
        range_ = (
            combined["load"].max() - combined["load"].min()
            if (combined["load"].max() - combined["load"].min()) != 0
            else np.nan
        )

    franks_skill_score = (mae(realised, basecase) - mae(realised, forecast)) / range_

    if np.isnan(franks_skill_score):
        return 0

    return franks_skill_score


def franks_skill_score_peaks(
    realised: pd.Series, forecast: pd.Series, basecase: pd.Series
) -> float:
    """Calculate Franks skill score on positive peaks."""
    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast, basecase], axis=1)

    range_ = (
        combined["load"].max() - combined["load"].min()
        if (combined["load"].max() - combined["load"].min()) != 0
        else np.nan
    )
    # Select 5 percent highest realised load values
    combined["highest"] = combined["load"][
        combined["load"] > combined["load"].quantile(0.95)
    ]
    combined = combined[np.invert(np.isnan(combined["highest"]))]

    # Calculate rMAE for the selected points
    franks_skill_score_highest = franks_skill_score(
        combined["load"],
        combined[forecast.name],
        combined[basecase.name],
        range_=range_,
    )

    if np.isnan(franks_skill_score_highest):
        return 0

    return franks_skill_score_highest


# Objective loss functions for quantile regression, from: https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7

# SPDX-FileCopyrightText: 2017 Nikolay Lysenko
#
# SPDX-License-Identifier: MIT


def xgb_quantile_eval(
    preds: np.ndarray, dmatrix: xgboost.DMatrix, quantile: float = 0.2
) -> str:
    """Customized evaluational metric that equals to quantile regression loss (also known as pinball loss).

    Quantile regression is regression that estimates a specified quantile of target's distribution conditional on given features.

    Args:
        preds: Predicted values
        dmatrix: xgboost.DMatrix of the input data.
        quantile: Target quantile.

    Returns:
        Loss information


    # See also:
    https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7

    """
    labels = dmatrix.get_label()
    return (
        "q{}_loss".format(quantile),
        np.nanmean(
            (preds >= labels) * (1 - quantile) * (preds - labels)
            + (preds < labels) * quantile * (labels - preds)
        ),
    )


def xgb_quantile_obj(
    preds: np.ndarray, dmatrix: xgboost.DMatrix, quantile: float = 0.2
) -> tuple[np.ndarray, np.ndarray]:
    """Quantile regression objective fucntion.

    Computes first-order derivative of quantile regression loss and a non-degenerate substitute for second-order
    derivative.

    Substitute is returned instead of zeros, because XGBoost requires non-zero second-order derivatives. See
    this page: https://github.com/dmlc/xgboost/issues/1825 to see why it is possible to use this trick. However, be sure
    that hyperparameter named `max_delta_step` is small enough to satisfy:``0.5 * max_delta_step <= min(quantile, 1 - quantile)``.

    Args:
        preds: numpy.ndarray
        dmatrix: xgboost.DMatrix
        quantile: float

    Returns:
        Gradient and Hessian

    # See also:
    https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7

    Reasoning for the hessian:
    https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7#gistcomment-2322558

    """
    try:
        assert 0 <= quantile <= 1
    except AssertionError:
        raise ValueError("Quantile value must be float between 0 and 1.")

    labels = dmatrix.get_label()
    errors = preds - labels

    left_mask = errors < 0
    right_mask = errors > 0

    # The factor `* errors` is different from the original implementation, however
    # this addition makes the objective function scalable with the size of the error.
    # This solves issues with regression on large (>100) input data.
    grad = (quantile * left_mask + (1 - quantile) * right_mask) * errors
    hess = np.ones_like(preds)

    return grad, hess



# SPDX-FileCopyrightText: 2017-2023 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the SARIMAX regressor wrapper around statsmodels implementation."""
import numpy as np
import pandas as pd
import statsmodels.api as sm

from sklearn.metrics import r2_score
from sklearn.model_selection import TimeSeriesSplit
from openstef.model.regressors.regressor import OpenstfRegressor


class ARIMAOpenstfRegressor(OpenstfRegressor):
    """Wrapper around statmodels implementation of (S)ARIMA(X) model.

    The fit of an ARIMA statsmodels produces a result object which is used to perform the various computations around forecasting.
    (see https://www.statsmodels.org/dev/generated/statsmodels.tsa.arima.model.ARIMAResults.html)

    To make a prediction, it needs to update the result object's historic data,
    ie the past values of the target/endogenous data and the features/exogenous data,
    applying the fitted parameters to these new data unrelated to the original training data.
    This update can be performed by the method `update_historic_data`.

    In the following code, we use interchangeably the statmodels and scikit-learn terminology for the variables:
        - the features 'x' is equivalent to the exogenous data: 'exog' for short.
        - the target 'y' is equivalent to the endogenous data: 'endog' for short.
    More information here https://www.statsmodels.org/stable/endog_exog.html.

    """

    def __init__(
        self,
        backtest_max_horizon=1440,
        order=(0, 0, 0),
        seasonal_order=(0, 0, 0, 0),
        trend=None,
    ):
        self.backtest_max_horizon = backtest_max_horizon
        self.order = order
        self.seasonal_order = seasonal_order
        self.trend = trend

    def fit(self, x, y, **kwargs):
        dates = x.index
        self.model_ = sm.tsa.arima.ARIMA(
            endog=y,
            exog=x,
            dates=dates,
            order=self.order,
            seasonal_order=self.seasonal_order,
            trend=self.trend,
        )
        self.results_ = self.model_.fit()
        self.feature_in_names_ = list(x.columns)
        return self

    def update_historic_data(self, x_past, y_past):
        """Apply the fitted parameters to new data unrelated to the original training data. It's a side-effect.

        Creates a new result object using the current fitted parameters,
        applied to a completely new dataset that is assumed to be unrelated to the model’s original data.
        The new results can then be used for analysis or forecasting.
        It should be used before forecasting, to wedge the historic data just before the first forecast timestamp,
        with:
            - New observations from the modeled time-series process.
            - New observations of exogenous regressors.

        Parameters
        ----------
        x_past : pd.DataFrame
            The exogenous (features) data.
        y_past : pd.DataFrame
            The endogenous (target) data.

        """
        self.results_ = self.results_.apply(
            endog=y_past,
            exog=x_past,
        )

    def predict_quantile(self, start, end, exog, quantile):
        """Quantile prediction.

        It relies on the parameters' confidence intervals.

        Parameters
        ----------
        start : int, str, or datetime, optional
            Zero-indexed observation number at which to start forecasting, i.e.,
            the first forecast is start. Can also be a date string to parse or a datetime type.
            Default is the the zeroth observation.
        end : int, str, or datetime, optional
            Zero-indexed observation number at which to end forecasting, i.e.,
            the last forecast is end. Can also be a date string to parse or a datetime type.
            However, if the dates index does not have a fixed frequency,
            end must be an integer index if you want out of sample prediction.
            Default is the last observation in the sample.
        exog : pd.DataFrame
            Exogenous data (features).
        quantile : float
            The quantile for the confidence interval.

        Returns
        -------
        pd.Serie
            The quantile prediction.

        """
        alpha = quantile
        idx = 0
        if quantile > 0.5:
            alpha = 1 - quantile
            idx = 1
        return (
            self.results_.get_prediction(start, end, exog=exog)
            .conf_int(alpha=alpha)
            .iloc[:, idx]
        )

    def predict(self, x, quantile: float = 0.5, **kwargs):
        start = x.iloc[0].name
        end = x.iloc[-1].name
        predictions = self.results_.predict(start, end, exog=x).to_numpy()
        if quantile != 0.5:
            predictions = self.predict_quantile(start, end, exog=x, quantile=quantile)
        return predictions

    def set_feature_importance(self):
        """Because report needs 'weight' and 'gain' as importance metrics, we set the values to these names.

        - 'weight' is corresponding to the coefficients values
        - 'gain' is corresponding to the pvalue for the nullity test of each coefficient

        """
        importances = pd.DataFrame(
            {"weight": self.results_.params, "gain": self.results_.pvalues}
        )
        return importances

    @property
    def feature_names(self):
        """The names of he features used to train the model."""
        return self.feature_in_names_

    @property
    def can_predict_quantiles(self):
        """Indicates wether this model can make quantile predictions."""
        return True

    def score(self, x, y):
        """Compute R2 score with backtesting strategy.

        The backtest  is performed by the Time Series cross-validator of scikit-learn which
        returns first k folds as train set and the (k+1)th fold as test set in the kth split.
        (see https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.TimeSeriesSplit.html)

        It needs to update the historic data with (x_past, y_past) for each split.

        """
        ys_true = []
        ys_pred = []

        # Build the cross-validator
        freq = pd.infer_freq(x.index)
        if not (freq[0].isdigit()):
            freq = f"1{freq}"
        max_horizon_delta = pd.Timedelta(self.backtest_max_horizon, "minutes")
        freq_delta = pd.Timedelta(freq)
        test_size = max_horizon_delta // freq_delta
        n_splits = (x.shape[0] // test_size) - 1
        time_series_cross_validator = TimeSeriesSplit(
            n_splits=n_splits, test_size=test_size
        )

        # Backtesting
        for apply_index, test_index in time_series_cross_validator.split(x):
            # Update the historic data to the current split (ie the k first folds)
            updated_results = self.results_.apply(
                y.iloc[apply_index], x.iloc[apply_index]
            )

            # The (k+1)th fold as the test data
            x_test, y_true_test = x.iloc[test_index], y.iloc[test_index]
            start_test = x_test.iloc[0].name
            end_test = x_test.iloc[-1].name

            # Compute and gather the predictions
            y_pred_test = updated_results.predict(
                start=start_test, end=end_test, exog=x_test
            )
            ys_true.append(y_true_test)
            ys_pred.append(y_pred_test)

        ys_true = np.concatenate(ys_true)
        ys_pred = np.concatenate(ys_pred)
        return r2_score(ys_true, ys_pred)



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains all holiday related features."""
from datetime import datetime, timedelta

import holidays
import numpy as np
import pandas as pd

from openstef import PROJECT_ROOT

HOLIDAY_CSV_PATH: str = (
    PROJECT_ROOT / "openstef" / "data" / "dutch_holidays_2020-2022.csv"
)


def generate_holiday_feature_functions(
    country: str = "NL",
    years: list = None,
    path_to_school_holidays_csv: str = HOLIDAY_CSV_PATH,
) -> dict:
    """Generates functions for creating holiday feature.

    This improves forecast accuracy. Examples of features that are
    added are: 2020-01-01 is 'Nieuwjaarsdag'.

        2022-12-24 - 2023-01-08 is the 'Kerstvakantie'
        2022-10-15 - 2022-10-23 is the 'HerfstvakantieNoord'


    The holidays are based on a manually generated csv file.
    The information is collected using:
    https://www.schoolvakanties-nederland.nl/ and the python holiday function
    The official following official ducth holidays are included untill 2023:
        - Kerstvakantie
        - Meivakantie
        - Herstvakantie
        - Bouwvak
        - Zomervakantie
        - Voorjaarsvakantie

        - Nieuwjaarsdag
        - Pasen
        - Koningsdag
        - Hemelvaart
        - Pinksteren
        - Kerst

    The 'Brugdagen' are updated untill dec 2020. (Generated using agenda)

    Args:
        country: Country for which to create holiday features.
        years: years for which to create holiday features.
        path_to_school_holidays_csv: Filepath to csv with school holidays.

    Returns:
        Dictionary with functions that check if a given date is a holiday, keys
        consist of "Is" + the_name_of_the_holiday_to_be_checked

    """
    if years is None:
        now = datetime.now()
        years = [now.year - 1, now.year]

    country_holidays = holidays.country_holidays(country, years=years)

    # Make holiday function dict
    holiday_functions = {}
    # Add check function that includes all holidays of the provided csv
    holiday_functions.update(
        {
            "is_national_holiday": lambda x: np.isin(
                x.index.date, np.array(list(country_holidays))
            )
        }
    )
    # Define empty list to keep track of bridgedays
    bridge_days = []
    # Loop over list of holidays names
    for date, holiday_name in sorted(country_holidays.items()):
        # Define function explicitely to mitigate 'late binding' problem
        def make_holiday_func(requested_date):
            return lambda x: np.isin(x.index.date, np.array([requested_date]))

        # Create lag function for each holiday
        holiday_functions.update(
            {"is_" + holiday_name.replace(" ", "_").lower(): make_holiday_func(date)}
        )

        # Check for bridge day
        holiday_functions, bridge_days = check_for_bridge_day(
            date, holiday_name, country, years, holiday_functions, bridge_days
        )

    # Add feature function that includes all bridgedays
    holiday_functions.update(
        {"is_bridgeday": lambda x: np.isin(x.index.date, np.array(list(bridge_days)))}
    )

    # Manully generated csv including all dutch schoolholidays for different regions
    df_holidays = pd.read_csv(path_to_school_holidays_csv, index_col=None)
    df_holidays["datum"] = pd.to_datetime(df_holidays.datum).apply(lambda x: x.date())

    # Add check function that includes all holidays of the provided csv
    holiday_functions.update(
        {"is_schoolholiday": lambda x: np.isin(x.index.date, df_holidays.datum.values)}
    )

    # Loop over list of holidays names
    for holiday_name in list(set(df_holidays.name)):
        # Define function explicitely to mitigate 'late binding' problem
        def make_holiday_func(holidayname=holiday_name):
            return lambda x: np.isin(
                x.index.date, df_holidays.datum[df_holidays.name == holidayname].values
            )

        # Create lag function for each holiday
        holiday_functions.update(
            {
                "is_"
                + holiday_name.replace(" ", "_").lower(): make_holiday_func(
                    holidayname=holiday_name
                )
            }
        )

    return holiday_functions


# Check for bridgedays
def check_for_bridge_day(
    date: datetime,
    holiday_name: str,
    country: str,
    years: list,
    holiday_functions: dict,
    bridge_days: list,
) -> tuple[dict, list]:
    """Checks for bridgedays associated to a specific holiday with date (date).

    Any found bridgedays are appende dto the bridgedays list. Also a specific feature
    function for the bridgeday is added to the general holidayfuncitons dictionary.

    Args:
        date: Date of holiday to check for associated bridgedays.
        holiday_name: Name of the holiday.
        country: Country for which to detect the bridgedays.
        years: List of years for which to detect bridgedays.
        holiday_functions: Dictionary to which the featurefunction has to be appended to in case of a bridgeday.
        bridge_days: List of bridgedays to which any found bridgedays have to be appended.

    Returns:
        - Dict with holiday feature functions
        - List of bridgedays

    """
    country_holidays = holidays.country_holidays(country, years=years)

    # Define function explicitely to mitigate 'late binding' problem
    def make_holiday_func(requested_date):
        return lambda x: np.isin(x.index.date, np.array([requested_date]))

    # Looking forward: If day after tomorow is a national holiday or
    # a saturday check if tomorow is not a national holiday

    is_saturday_in_two_days = (date + timedelta(days=2)).weekday() == 5
    is_holiday_in_two_days = (date + timedelta(days=2)) in country_holidays

    is_holiday_tommorow = (date + timedelta(days=1)) in country_holidays
    is_weekend_tommorrow = (date + timedelta(days=1)).weekday() in [5, 6]

    if (is_holiday_in_two_days or is_saturday_in_two_days) and (
        not is_holiday_tommorow and not is_weekend_tommorrow
    ):

        # Create feature function for each holiday
        holiday_functions.update(
            {
                "is_bridgeday"
                + holiday_name.replace(" ", "_").lower(): make_holiday_func(
                    (date + timedelta(days=1))
                )
            }
        )
        bridge_days.append((date + timedelta(days=1)))

    # Looking backward: If day before yesterday is a national holiday
    # or a sunday check if yesterday is a national holiday
    is_saturday_two_days_ago = (date - timedelta(days=2)).weekday() == 6
    is_holiday_two_days_ago = (date - timedelta(days=2)) in country_holidays
    is_holiday_yesterday = (date - timedelta(days=1)) in country_holidays
    is_weekend_yesterday = (date - timedelta(days=1)).weekday() in [5, 6]

    if (is_saturday_two_days_ago or is_holiday_two_days_ago) and (
        not is_holiday_yesterday and not is_weekend_yesterday
    ):

        # Create featurefunction for the bridge function
        holiday_functions.update(
            {
                "is_bridgeday"
                + holiday_name.replace(" ", "_").lower(): make_holiday_func(
                    (date - timedelta(days=1))
                )
            }
        )
        bridge_days.append((date - timedelta(days=1)))

    return holiday_functions, bridge_days



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""This module contains the CRON job that is periodically executed to make prognoses of solar features.

These features are usefull for splitting the load
in solar and wind contributions.
This is achieved by carrying out the folowing steps:
  1. Get the wind and solar reference data for the specific location of the
     customer
  2. Get the TDCV (Typical Domestic Consumption Values) data
  3. Fit a linear combination of above time series to the historic load data to
     determine the contributions of each energy source.
  4. Write the resulting coeficients to the SQL database.

Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::

        $ python split_forecast.py

"""
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import scipy.optimize
import structlog

import openstef.monitoring.teams as monitoring
from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.enums import MLModelType
from openstef.tasks.utils.predictionjobloop import PredictionJobLoop
from openstef.tasks.utils.taskcontext import TaskContext

COEF_MAX_FRACTION_DIFF = 0.3


def main(config=None, database=None):
    taskname = Path(__file__).name.replace(".py", "")

    if database is None or config is None:
        raise RuntimeError(
            "Please specifiy a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    with TaskContext(taskname, config, database) as context:
        model_type = [ml.value for ml in MLModelType]

        PredictionJobLoop(
            context,
            model_type=model_type,
        ).map(split_forecast_task, context)


def split_forecast_task(
    pj: PredictionJobDataClass,
    context: TaskContext,
) -> pd.DataFrame:
    """Function that caries out the energy splitting for a specific prediction job with id pid.

    Args:
        pid: Prediction job id

    Returns:
        Energy splitting coefficients.

    """
    logger = structlog.get_logger(__name__)

    logger.info("Start splitting energy", pid=pj["id"])

    # Get input for splitting
    input_split_function = context.database.get_input_energy_splitting(pj)

    # Old split method;
    # find_components() gives two things:
    # - the split components (load, solar, wind, consumption, Inschatting (=sum of others) )
    # - coefdict: coefficients of each component; these are not yet an output of Dazls. Lets discuss with JM if we want that

    # Carry out the splitting
    components, coefdict = find_components(input_split_function)

    # Calculate mean absolute error (MAE)
    # TODO: use a standard metric function for this
    error = components[["load", "Inschatting"]].diff(axis=1).iloc[:, 1]
    mae = error.abs().mean()
    coefdict.update({"MAE": mae})
    coefsdf = convert_coefdict_to_coefsdf(pj, input_split_function, coefdict)

    # Get the coefs of previous runs and check if new coefs are valid
    last_coefsdict = context.database.get_energy_split_coefs(pj)
    last_coefsdf = convert_coefdict_to_coefsdf(pj, input_split_function, last_coefsdict)
    invalid_coefs = determine_invalid_coefs(coefsdf, last_coefsdf)
    if not invalid_coefs.empty:
        # If coefs not valid, do not update the coefs in the db and send teams
        # message that something strange is happening
        monitoring.post_teams(
            f"New splitting coefficient(s) for pid **{pj['id']}** deviate strongly "
            "from previously stored coefficients.",
            url=context.config.teams_monitoring_url,
            invalid_coefficients=invalid_coefs,
            coefficients_df=coefsdf,
        )
        # Use the last known coefficients for further processing
        return last_coefsdf
    else:
        # Save Results
        context.database.write_energy_splitting_coefficients(
            coefsdf, if_exists="append"
        )
        logger.info(
            "Succesfully wrote energy split coefficients to database", pid=pj["id"]
        )
        return coefsdf


def determine_invalid_coefs(
    new_coefs: pd.DataFrame, last_coefs: pd.DataFrame
) -> pd.DataFrame:
    """Determine which new coefficients are valid and return them.

    Args:
        new_coefs: df of new coefficients for standard load
            profiles (i.e. wind, solar, household)
        last_coefs: df of last coefficients for standard load
            profiles (i.e. wind, solar, household)

    Returns:
        Dataframe with invalid coefficients

    """
    merged_coefs = pd.merge(
        last_coefs, new_coefs, on="coef_name", how="left", suffixes=["_last", "_new"]
    )
    # calculate difference between new and last coefficients, if no new
    # coefficient, set difference to inf
    # If coefficient name is not present in new coefficients list, fail. If coefficient
    # name is not present in last coefficients list, add it.
    merged_coefs["difference"] = (
        (merged_coefs.coef_value_last - merged_coefs.coef_value_new)
        .abs()
        .fillna(np.inf)
    )
    # Check if the absolute difference between last coefficients and new coefficients
    # is more than COEF_MAX_FRACTION_DIFF x absolute value of last coefficient
    invalid_coefs = merged_coefs[
        merged_coefs.difference
        > (COEF_MAX_FRACTION_DIFF * merged_coefs.coef_value_last).abs()
    ]
    return invalid_coefs


def convert_coefdict_to_coefsdf(
    pj: PredictionJobDataClass, input_split_function: pd.DataFrame, coefdict: dict
) -> pd.DataFrame:
    """Convert dictionary of coefficients to dataframe with additional data for db storage.

    Args:
        pj: prediction job
        input_split_function: df of columns of standard load profiles,
            i.e. wind, solar, household
        coefdict: dict of coefficient per standard load profile

    Returns:
        DataFrame of coefficients to insert in sql

    """
    #
    sql_column_labels = ["pid", "date_start", "date_end", "created"]
    sql_colum_values = [
        pj["id"],
        input_split_function.index.min().date(),
        input_split_function.index.max().date(),
        datetime.utcnow(),
    ]
    coefsdf = pd.DataFrame(
        {"coef_name": list(coefdict.keys()), "coef_value": list(coefdict.values())}
    )
    for i, column in enumerate(sql_column_labels):
        coefsdf[column] = sql_colum_values[i]

    return coefsdf


def find_components(
    df: pd.DataFrame, zero_bound: bool = True
) -> tuple[pd.DataFrame, dict]:
    """Function that does the actual energy splitting.

    Args:
        df: Input data. The dataframe should contain these columns
            in exactly this order: [load, wind_ref, pv_ref, mulitple tdcv colums]
        zero_bound: If zero_bound is True coefficients can't be negative.

    Returns:
        tuple:
            - DataFrame containing the wind and solar components
            - Dict with the coefficients that result from the fitting

    """
    # Define function to fit
    def weighted_sum(x, *args):
        if len(x) != len(args):
            raise ValueError("Length of args should match len of x")
        weights = np.array([v for v in args])
        return np.dot(x.T, weights)

    load = df.iloc[:, 0]
    wind_ref = df.iloc[:, 1]
    pv_ref = df.iloc[:, 2]

    # Define scaler
    nedu_scaler = (load.max() - load.min()) / 10

    # Come up with inital guess for the fitting
    p_wind_guess = 1.0
    ppv_guess = 1.0
    p0 = [p_wind_guess, ppv_guess] + (len(df.columns) - 3) * [nedu_scaler]

    # Define fitting bounds
    if zero_bound:
        bounds = (0, "inf")
    else:
        bounds = ("-inf", "inf")

    # Carry out fitting
    # See https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html # noqa
    coefs, cov = scipy.optimize.curve_fit(
        weighted_sum,
        xdata=df.iloc[:, 1:].values.T,
        ydata=load.values,
        p0=p0,
        bounds=bounds,
        method="trf",
    )

    # Set 'almost zero' to zero
    coefs[coefs < 0.1] = 0

    # Reconstuct historical load
    hist = weighted_sum(df.iloc[:, 1:].values.T, *coefs)
    histp0 = weighted_sum(df.iloc[:, 1:].values.T, *p0)

    # Make a nice dataframe to return the components
    components = df.iloc[:, [0]].copy()
    components["Inschatting"] = hist.T
    components["p0"] = histp0.T
    components["Windopwek"] = wind_ref * coefs[0]
    components["Zonne-opwek"] = pv_ref * coefs[1]
    components["StandaardVerbruik"] = (df.iloc[:, 3:] * coefs[2:]).sum(axis=1)
    components["Residu"] = -1 * components.iloc[:, 0:2].diff(axis=1).iloc[:, 1]

    # Make nice dictinary to return coefficents
    coefdict = {name: value for name, value in zip(df.columns[1:], coefs)}

    # Return result
    return components, coefdict


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the CRON job that is periodically executed to make the components prognoses.

This code assumes trained models are available from the persistent storage.
If these are not available run model_train.py to train all models.
To provide the prognoses the following steps are carried out:
  1. Get historic training data (TDCV, Load, Weather and APX price data)
  2. Apply features
  3. Load model
  4. Make component prediction
  5. Write prediction to the database
  6. Send Teams message if something goes wrong

Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::

        $ python create_components_forecast.py

"""
from datetime import datetime, timedelta, timezone
from pathlib import Path

import structlog
import pandas as pd

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.enums import MLModelType
from openstef.exceptions import ComponentForecastTooShortHorizonError
from openstef.pipeline.create_component_forecast import (
    create_components_forecast_pipeline,
)
from openstef.tasks.utils.predictionjobloop import PredictionJobLoop
from openstef.tasks.utils.taskcontext import TaskContext

T_BEHIND_DAYS = 0
T_AHEAD_DAYS = 3


def create_components_forecast_task(
    pj: PredictionJobDataClass, context: TaskContext
) -> None:
    """Top level task that creates a components forecast.

    On this task level all database and context manager dependencies are resolved.

    Args:
        pj: Prediction job
        context: Contect object that holds a config manager and a database connection

    """
    logger = structlog.get_logger(__name__)
    if pj["train_components"] == 0:
        context.logger.info(
            "Skip prediction job", train_components=pj["train_components"]
        )
        return

    # Define datetime range for input data
    datetime_start = datetime.utcnow() - timedelta(days=T_BEHIND_DAYS)
    datetime_end = datetime.utcnow() + timedelta(days=T_AHEAD_DAYS)

    logger.info(
        "Get predicted load", datetime_start=datetime_start, datetime_end=datetime_end
    )
    # Get most recent load forecast as input_data,
    # we use a regular forecast as input point for creating component forecasts
    input_data = context.database.get_predicted_load(
        pj, start_time=datetime_start, end_time=datetime_end
    )
    # Check if input_data is not empty
    if len(input_data) == 0:
        logger.warning(f"No forecast found. Skipping pid", pid=pj["id"])
        return

    logger.info("retrieving weather data")
    # TODO make openstef_dbc function to retrieve inputdata for component forecast in one call,
    #  this will make this function much shorter
    # Get required weather data
    weather_data = context.database.get_weather_data(
        [pj["lat"], pj["lon"]],
        [
            "radiation",
            "windspeed_100m",
        ],  # These variables are used when determing the splitting coeficients, and should therefore be reused when making the component forcasts.
        datetime_start=datetime_start,
        datetime_end=datetime_end,
    )

    # Make forecast for the demand, wind and pv components
    forecasts = create_components_forecast_pipeline(pj, input_data, weather_data)

    ## Perform sanity check on index
    if not isinstance(forecasts.index, pd.core.indexes.datetimes.DatetimeIndex):
        raise ValueError(
            f"Index is not datetime. Received forecasts:{forecasts.head()}"
        )

    # save forecast to database #######################################################
    context.database.write_forecast(forecasts)
    logger.debug("Written forecast to database")

    # Check if forecast was complete enough, otherwise raise exception
    if forecasts.index.max() < datetime.utcnow().replace(
        tzinfo=timezone.utc
    ) + timedelta(hours=30):

        # Check which input data is missing the most.
        # Do this by counting the NANs for (load)forecast, radiation and windspeed
        max_index = forecasts.index.max()
        n_nas = dict(
            nans_load_forecast=input_data.loc[max_index:, "forecast"].isna().sum(),
            nans_radiation=weather_data.loc[max_index:, "radiation"].isna().sum(),
            nans_windspeed_100m=weather_data.loc[max_index:, "windspeed_100m"]
            .isna()
            .sum(),
        )
        max_na = max(n_nas, key=n_nas.get)

        raise ComponentForecastTooShortHorizonError(
            f"Could not make component forecast for two days ahead, probably input data is missing, {max_na}: {n_nas[max_na]}"
        )


def main(config: object = None, database: object = None):
    taskname = Path(__file__).name.replace(".py", "")

    if database is None or config is None:
        raise RuntimeError(
            "Please specifiy a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    with TaskContext(taskname, config, database) as context:

        model_type = [ml.value for ml in MLModelType]

        PredictionJobLoop(
            context,
            model_type=model_type,
        ).map(create_components_forecast_task, context)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import joblib
import pandas as pd
import structlog

import openstef.postprocessing.postprocessing as postprocessing
from openstef import PROJECT_ROOT
from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.enums import ForecastType
from openstef.model.regressors.dazls import Dazls

# Set the path for the Dazls stored model
DAZLS_STORED = PROJECT_ROOT / "openstef" / "data" / "dazls_stored.sav"


def create_input(
    pj: PredictionJobDataClass, input_data: pd.DataFrame, weather_data: pd.DataFrame
) -> pd.DataFrame:
    """This function prepares the input data.

    This data will be used for the Dazls model prediction, so they will be
    according Dazls model requirements.

    Args:
        pj: Prediction job
        input_data: Input forecast for the components forecast.
        weather_data: Weather data with 'radiation' and 'windspeed_100m' columns

    Returns:
        It outputs a dataframe which will be used for the Dazls prediction function.

    """
    # Prepare raw input data
    input_df = (
        weather_data[["radiation", "windspeed_100m"]]
        .merge(
            input_data[["forecast"]].rename(columns={"forecast": "total_substation"}),
            how="inner",
            right_index=True,
            left_index=True,
        )
        .dropna()
    )
    # Add additional features
    input_df["lat"] = pj["lat"]
    input_df["lon"] = pj["lon"]

    input_df["solar_on"] = 1
    input_df["wind_on"] = 1
    input_df["hour"] = input_df.index.hour
    input_df["minute"] = input_df.index.minute

    input_df["var0"] = input_df["total_substation"].var()
    input_df["var1"] = input_df["radiation"].var()
    input_df["var2"] = input_df["windspeed_100m"].var()

    input_df["sem0"] = input_df["total_substation"].sem()
    input_df["sem1"] = input_df["radiation"].sem()

    return input_df


def create_components_forecast_pipeline(
    pj: PredictionJobDataClass, input_data: pd.DataFrame, weather_data: pd.DataFrame
) -> pd.DataFrame:
    """Pipeline for creating a component forecast using Dazls prediction model.

    Args:
        pj: Prediction job
        input_data: Input forecast for the components forecast.
        weather_data: Weather data with 'radiation' and 'windspeed_100m' columns

    Returns:
        DataFrame with component forecasts. The dataframe contains these columns;
        "forecast_wind_on_shore",
        "forecast_solar",
        "forecast_other",
        "pid",
        "customer",
        "description",
        "type",
        "algtype"

    """
    logger = structlog.get_logger(__name__)
    logger.info("Make components prediction", pid=pj["id"])

    # Make component forecasts
    try:
        input_data = create_input(pj, input_data, weather_data)

        # Save and load the model as .sav file
        # For the code contact: korte.termijn.prognoses@alliander.com
        dazls_model: Dazls = joblib.load(DAZLS_STORED)

        # Use the predict function of Dazls model
        # As input data we use the input_data function which takes into consideration what we want as an input for the forecast and what Dazls can accept as an input
        forecasts = dazls_model.predict(x=input_data)

        # Set the columns for the output forecast dataframe
        forecasts = pd.DataFrame(
            forecasts,
            columns=["forecast_wind_on_shore", "forecast_solar"],
            index=input_data.index,
        )

        # Make post-processed forecasts for solar and wind power
        # These forecasts are respectively for the components: "forecast_solar" and "forecast_wind_on_shore"
        # The outcome forecasts are added in the "forecasts" DataFrame we created above
        forecasts["forecast_solar"] = postprocessing.post_process_wind_solar(
            forecasts["forecast_solar"], forecast_type=ForecastType.SOLAR
        )
        forecasts["forecast_wind_on_shore"] = postprocessing.post_process_wind_solar(
            forecasts["forecast_wind_on_shore"], forecast_type=ForecastType.WIND
        )

        # Make forecast for the component: "forecast_other"
        forecasts["forecast_other"] = (
            input_data["total_substation"]
            - forecasts["forecast_solar"]
            - forecasts["forecast_wind_on_shore"]
        )
    except Exception as e:
        # In case something goes wrong we fall back on aan empty dataframe
        logger.warning(
            f"Could not make component forecasts: {e}, falling back on series of"
            " zeros!",
            exc_info=e,
        )
        forecasts = pd.DataFrame()

    # Prepare for output
    # Add more prediction properties to the forecast ("pid","customer","description","type","algtype)
    forecasts = postprocessing.add_prediction_job_properties_to_forecast(
        pj, forecasts, algorithm_type="component"
    )
    return forecasts



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module defines the DAZL model."""
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.neighbors import KNeighborsRegressor
from sklearn.preprocessing import MinMaxScaler
from sklearn.utils import shuffle


class Dazls(BaseEstimator):
    """DAZLS model.

    The model carries out wind and solar power prediction for unseen target substations using training data from
    other substations with known components.

    Any data-driven model can be plugged and used as the base for the domain and the adaptation model.

    For a full reference, see:
    Teng, S.Y., van Nooten, C. C., van Doorn, J.M., Ottenbros, A., Huijbregts, M., Jansen, J.J.
    Improving Near Real-Time Predictions of Renewable Electricity Production at Substation Level (Submitted)

    """

    def __init__(self):
        """Initialize DAZL model."""
        self.__name__ = "DAZLS"
        self.domain_model_scaler = MinMaxScaler(clip=True)
        self.adaptation_model_scaler = MinMaxScaler(clip=True)
        self.target_scaler = MinMaxScaler(clip=True)
        self.domain_model = KNeighborsRegressor(n_neighbors=20, weights="uniform")
        self.adaptation_model = KNeighborsRegressor(n_neighbors=20, weights="uniform")

        # The input columns for the domain and adaptation models (with description)
        self.domain_model_input_columns = [
            "radiation",  # Weather parameter
            "windspeed_100m",  # Weather parameter
            "total_substation",  # Substation's measured total load
            "lat",  # Latitude
            "lon",  # Longitude
            "solar_on",  # Solar installed on substation: yes=1, no=0
            "wind_on",  # Wind installed on substation: yes=1, no=0
            "hour",  # Hour of the day
            "minute",  # Minute of the hour
            "var0",  # Variance of the total load
            "var1",  # Variance of the total pv load (only available for calibration substations)
            "var2",  # Variance of the total wind load (only available for calibration substations)
            "sem0",  # Standard Error of the Mean of the total load
            "sem1",  # Standard Error of the Mean of the total PV load (only available for calibration substations)
        ]
        self.adaptation_model_input_columns = [
            "total_substation",
            "lat",
            "lon",
            "solar_on",
            "wind_on",
            "hour",
            "minute",
            "var0",
            "var1",
            "var2",
            "sem0",
            "sem1",
        ]
        self.target_columns = ["total_wind_part", "total_solar_part"]

    def fit(self, features, target):
        """Fit the model.

        In this function we scale the input of the domain and adaptation models of the DAZLS MODEL. Then we fit the
        two models. We separate the features into domain_model_input, adaptation_model_input and target, and we use them
        for the fitting and the training of the models.

        Args:
            features: inputs for domain and adaptation model (domain_model_input, adaptation_model_input)
            target: the expected output (y_train)

        """
        x, x2, y = (
            features.loc[:, self.domain_model_input_columns],
            features.loc[:, self.adaptation_model_input_columns],
            target.loc[:, self.target_columns],
        )
        domain_model_input, adaptation_model_input, y_train = shuffle(
            x, x2, y, random_state=999
        )  # just shuffling

        self.domain_model_scaler.fit(domain_model_input)
        self.adaptation_model_scaler.fit(adaptation_model_input)
        self.target_scaler.fit(y_train)
        domain_model_input = self.domain_model_scaler.transform(domain_model_input)
        adaptation_model_input = self.adaptation_model_scaler.transform(
            adaptation_model_input
        )
        y_train = self.target_scaler.transform(y_train)

        self.domain_model.fit(domain_model_input, y_train)
        domain_model_pred = self.domain_model.predict(domain_model_input)
        adaptation_model_input = np.concatenate(
            (adaptation_model_input, domain_model_pred), axis=1
        )
        self.adaptation_model.fit(adaptation_model_input, y_train)

    def predict(self, x: np.array):
        """Make a prediction.

        For the prediction we use the test data x. We use domain_model_input_columns and
        adaptation_model_input_columns to separate x in test data for domain model and adaptation model respectively.

        Args:
            x: domain_model_test_data, adaptation_model_test_data
            prediction: The output prediction after both models.

        """
        domain_model_test_data, adaptation_model_test_data = (
            x.loc[:, self.domain_model_input_columns],
            x.loc[:, self.adaptation_model_input_columns],
        )
        # Rescale test data for both models (if required)
        domain_model_test_data_scaled = self.domain_model_scaler.transform(
            domain_model_test_data
        )
        adaptation_model_test_data_scaled = self.adaptation_model_scaler.transform(
            adaptation_model_test_data
        )
        # Use the scaled data to make domain_model_prediction
        domain_model_test_data_pred = self.domain_model.predict(
            domain_model_test_data_scaled
        )
        # Use the domain_model_prediction to make adaptation_model_prediction
        adaptation_model_test_data_pred = self.adaptation_model.predict(
            np.concatenate(
                [adaptation_model_test_data_scaled, domain_model_test_data_pred], axis=1
            )
        )
        # Rescale adaptation_model_prediction (if required)
        prediction = self.target_scaler.inverse_transform(
            adaptation_model_test_data_pred
        )
        return prediction

    def score(self, truth, prediction):
        """Evaluation of the prediction's output.

        Args:
            truth: real values
            prediction: predicted values

        Returns:
            RMSE and R2 scores

        """
        rmse = (mean_squared_error(truth, prediction)) ** 0.5
        r2_score_value = r2_score(truth, prediction)
        return rmse, r2_score_value



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import json
import os
import shutil
from datetime import datetime
from json import JSONDecodeError
from typing import Optional, Union
from urllib.parse import unquote, urlparse

import mlflow
import numpy as np
import pandas as pd
import structlog
from mlflow.exceptions import MlflowException
from xgboost import XGBModel  # Temporary for backward compatibility

from openstef.data_classes.model_specifications import ModelSpecificationDataClass
from openstef.metrics.reporter import Report
from openstef.model.regressors.regressor import OpenstfRegressor


class MLflowSerializer:
    def __init__(self, mlflow_tracking_uri: str):
        self.logger = structlog.get_logger(self.__class__.__name__)
        mlflow.set_tracking_uri(mlflow_tracking_uri)
        self.logger.debug(f"MLflow tracking uri at init= {mlflow_tracking_uri}")

    def save_model(
        self,
        model: OpenstfRegressor,
        experiment_name: str,
        model_type: str,
        model_specs: ModelSpecificationDataClass,
        report: Report,
        phase: str = "training",
        **kwargs,
    ) -> None:
        """Save sklearn compatible model to MLFlow."""
        mlflow.set_experiment(experiment_name=experiment_name)
        with mlflow.start_run(run_name=experiment_name):
            self._log_model_with_mlflow(
                model=model,
                experiment_name=experiment_name,
                model_type=model_type,
                model_specs=model_specs,
                report=report,
                phase=phase,
                **kwargs,
            )
            self._log_figures_with_mlflow(report)

    def _log_model_with_mlflow(
        self,
        model: OpenstfRegressor,
        experiment_name: str,
        model_type: str,
        model_specs: ModelSpecificationDataClass,
        report: Report,
        phase: str,
        **kwargs,
    ) -> None:
        """Log model with MLflow.

        Note: **kwargs has extra information to be logged with mlflow

        """
        # Get previous run id
        models_df = self._find_models(
            experiment_name, max_results=1
        )  # returns latest model
        if not models_df.empty:
            previous_run_id = models_df["run_id"][
                0
            ]  # Use [0] to only get latest run id
        else:
            self.logger.info(
                "No previous model found in MLflow", experiment_name=experiment_name
            )
            previous_run_id = None

        # Set tags to the run, can be used to filter on the UI
        mlflow.set_tag("run_id", mlflow.active_run().info.run_id)
        mlflow.set_tag("phase", phase)  # phase can be Training or Hyperparameter_opt
        mlflow.set_tag("Previous_version_id", previous_run_id)
        mlflow.set_tag("model_type", model_type)
        mlflow.set_tag("prediction_job", experiment_name)

        # Add feature names, target, feature modules, metrics and params to the run
        mlflow.set_tag(
            "feature_names", model_specs.feature_names[1:]
        )  # feature names are 1+ columns
        mlflow.set_tag("target", model_specs.feature_names[0])  # target is first column
        mlflow.set_tag("feature_modules", model_specs.feature_modules)
        mlflow.log_metrics(report.metrics)
        model_specs.hyper_params.update(model.get_params())
        # TODO: Remove this hardcoded hyper params fix with loop after fix by mlflow
        # https://github.com/mlflow/mlflow/issues/6384
        for key, value in model_specs.hyper_params.items():
            if value == "":
                model_specs.hyper_params[key] = " "
        mlflow.log_params(model_specs.hyper_params)

        # Process args
        for key, value in kwargs.items():
            if isinstance(value, dict):
                mlflow.log_dict(value, f"{key}.json")
            elif isinstance(value, str) or isinstance(value, int):
                mlflow.set_tag(key, value)
            else:
                self.logger.warning(
                    f"Couldn't log {key}, {type(key)} not supported",
                    experiment_name=experiment_name,
                )

        # Log the model to the run. Signature describes model input and output scheme
        mlflow.sklearn.log_model(
            sk_model=model, artifact_path="model", signature=report.signature
        )
        self.logger.info("Model saved with MLflow", experiment_name=experiment_name)

    def _log_figures_with_mlflow(self, report) -> None:
        """Log figures with MLflow in the artifact folder."""
        if report.feature_importance_figure is not None:
            mlflow.log_figure(
                report.feature_importance_figure, "figures/weight_plot.html"
            )
        for key, figure in report.data_series_figures.items():
            mlflow.log_figure(figure, f"figures/{key}.html")
        self.logger.info("Logged figures to MLflow.")

    def load_model(
        self,
        experiment_name: str,
    ) -> tuple[OpenstfRegressor, ModelSpecificationDataClass]:
        """Load sklearn compatible model from MLFlow.

        Args:
            experiment_name: Name of the experiment, often the id of the predition job.

        """
        try:
            models_df = self._find_models(
                experiment_name, max_results=1
            )  # return the latest finished run of the model
            if not models_df.empty:
                latest_run = models_df.iloc[0]  # Use .iloc[0] to only get latest run
            else:
                raise LookupError(
                    f"Model not found for experiment_name {experiment_name}. First"
                    " train a model!"
                )
            model_uri = self._get_model_uri(latest_run.artifact_uri)
            loaded_model = mlflow.sklearn.load_model(model_uri)
            loaded_model.age = self._determine_model_age_from_mlflow_run(latest_run)
            model_specs = self._get_model_specs(
                experiment_name, loaded_model, latest_run
            )
            loaded_model.path = unquote(
                urlparse(model_uri).path
            )  # Path without file:///
            self.logger.info("Model successfully loaded with MLflow")
            return loaded_model, model_specs
        except (AttributeError, MlflowException, OSError) as e:
            raise AttributeError(
                f"Model not found for experiment_name {experiment_name}. First train a"
                " model!"
            ) from e

    def get_model_age(
        self, experiment_name: str, hyperparameter_optimization_only: bool = False
    ) -> int:
        """Get model age of most recent model.

        Args:
            experiment_name: Name of the experiment, often the id of the predition job.
            hyperparameter_optimization_only: Set to true if only hyperparameters optimaisation events should be considered.

        """
        filter_string = "attribute.status = 'FINISHED'"
        if hyperparameter_optimization_only:
            filter_string += " AND tags.phase = 'Hyperparameter_opt'"
        models_df = self._find_models(
            experiment_name, max_results=1, filter_string=filter_string
        )
        if not models_df.empty:
            run = models_df.iloc[0]  # Use .iloc[0] to only get latest run
            return self._determine_model_age_from_mlflow_run(run)
        else:
            self.logger.info("No model found returning infinite model age!")
            return np.inf

    def _find_models(
        self,
        experiment_name: str,
        max_results: Optional[int] = 100,
        filter_string: str = "attribute.status = 'FINISHED'",
    ) -> pd.DataFrame:
        """Finds trained models for specific experiment_name sorted by age in descending order."""
        models_df = mlflow.search_runs(
            experiment_names=[experiment_name],
            max_results=max_results,
            filter_string=filter_string,
        )
        return models_df

    def _get_model_specs(
        self,
        experiment_name: str,
        loaded_model: OpenstfRegressor,
        latest_run: pd.Series,
    ) -> ModelSpecificationDataClass:
        """Get model specifications from existing model."""
        model_specs = ModelSpecificationDataClass(id=experiment_name)

        # Temporary fix for update of xgboost
        # new version requires some attributes that the old (stored) models don't have yet
        # see: https://stackoverflow.com/questions/71912084/attributeerror-xgbmodel-object-has-no-attribute-callbacks
        new_attrs = [
            "grow_policy",
            "max_bin",
            "eval_metric",
            "callbacks",
            "early_stopping_rounds",
            "max_cat_to_onehot",
            "max_leaves",
            "sampling_method",
        ]

        manual_additional_attrs = [
            "enable_categorical",
            "predictor",
        ]  # these ones are not mentioned in the stackoverflow post
        automatic_additional_attrs = [
            x
            for x in XGBModel._get_param_names()
            if x
            not in new_attrs + manual_additional_attrs + loaded_model._get_param_names()
        ]

        for attr in new_attrs + manual_additional_attrs + automatic_additional_attrs:
            setattr(loaded_model, attr, None)

        # This one is new is should be set to a specific value (https://xgboost.readthedocs.io/en/latest/python/python_api.html#module-xgboost.training)
        setattr(loaded_model, "missing", np.nan)
        setattr(loaded_model, "n_estimators", 100)

        # End temporary fix

        # get the parameters from old model, we insert these later into new model
        model_specs.hyper_params = loaded_model.get_params()
        # TODO: Remove this hardcoded hyper params fix with loop after fix by mlflow
        # https://github.com/mlflow/mlflow/issues/6384
        for key, value in model_specs.hyper_params.items():
            if value == " ":
                model_specs.hyper_params[key] = ""
        # get used feature names else use all feature names
        model_specs.feature_names = self._get_feature_names(
            experiment_name, latest_run, model_specs, loaded_model
        )
        # get feature_modules
        model_specs.feature_modules = self._get_feature_modules(
            experiment_name, latest_run, model_specs, loaded_model
        )
        return model_specs

    def _determine_model_age_from_mlflow_run(self, run: pd.Series) -> Union[int, float]:
        """Determines how many days ago a model is trained from the mlflow run."""
        try:
            model_datetime = run.end_time.to_pydatetime()
            model_datetime = model_datetime.replace(tzinfo=None)
            model_age_days = (datetime.utcnow() - model_datetime).days
        except Exception as e:
            self.logger.warning(
                "Could not get model age. Returning infinite age!", exception=str(e)
            )
            return np.inf  # Return fallback age
        return model_age_days

    def remove_old_models(
        self, experiment_name: str, max_n_models: int = 10, artifact_folder: str = None
    ):
        """Remove old models per experiment.

        Note: This functionality is not incorporated in MLFlow natively
        See also: https://github.com/mlflow/mlflow/issues/2152

        """
        if max_n_models < 1:
            raise ValueError(
                f"Max models to keep should be greater than 1! Received: {max_n_models}"
            )
        previous_runs = self._find_models(experiment_name=experiment_name)
        if len(previous_runs) > max_n_models:
            self.logger.debug(
                f"Going to delete old models. {len(previous_runs)} > {max_n_models}"
            )
            # Find run_ids of oldest runs
            runs_to_remove = previous_runs.sort_values(
                by="end_time", ascending=False
            ).loc[max_n_models:, :]
            for _, run in runs_to_remove.iterrows():
                self.logger.debug(
                    f"Going to remove run {run.run_id}, from {run.end_time}."
                )
                mlflow.delete_run(run.run_id)
                self.logger.debug("Removed run")

                # mlflow.delete_run marks it as deleted but does not delete it by itself
                if artifact_folder:  # Also try to remove artifact from disk.
                    artifact_filepath = (
                        f"{artifact_folder}/mlruns/{run.experiment_id}/{run.run_id}"
                    )
                    self.logger.debug(f"Removing artifact: {artifact_filepath}")
                    try:
                        shutil.rmtree(artifact_filepath)
                        self.logger.debug("Removed artifact")
                    except Exception as e:
                        self.logger.info(f"Failed removing artifacts: {e}")

    def _get_feature_names(
        self,
        experiment_name: str,
        latest_run: pd.Series,
        model_specs: ModelSpecificationDataClass,
        loaded_model: OpenstfRegressor,
    ) -> list:
        """Get the feature_names from MLflow or the old model."""
        error_message = "feature_names not loaded and using None, because it"
        try:
            model_specs.feature_names = json.loads(
                latest_run["tags.feature_names"].replace("'", '"')
            )
        except KeyError:
            self.logger.warning(
                f"{error_message} did not exist in run",
                experiment_name=experiment_name,
            )
        except AttributeError:
            self.logger.warning(
                f"{error_message} needs to be a string",
                experiment_name=experiment_name,
            )
        except JSONDecodeError:
            self.logger.warning(
                f"{error_message} needs to be a string of a list",
                experiment_name=experiment_name,
            )

        # if feature names is none, see if we can retrieve them from the old model
        if model_specs.feature_names is None:
            try:
                if loaded_model.feature_names is not None:
                    model_specs.feature_names = loaded_model.feature_names
                    self.logger.info(
                        "feature_names retrieved from old model with an attribute",
                        experiment_name=experiment_name,
                    )
            except AttributeError:
                self.logger.warning(
                    "feature_names not an attribute of the old model, using None ",
                    experiment_name=experiment_name,
                )
        return model_specs.feature_names

    def _get_feature_modules(
        self,
        experiment_name: str,
        latest_run: pd.Series,
        model_specs: ModelSpecificationDataClass,
        loaded_model: OpenstfRegressor,
    ) -> list:
        """Get the feature_modules from MLflow or the old model."""
        error_message = "feature_modules not loaded and using None, because it"
        try:
            model_specs.feature_modules = json.loads(
                latest_run["tags.feature_modules"].replace("'", '"')
            )

        except KeyError:
            self.logger.warning(
                f"{error_message} did not exist in run",
                experiment_name=experiment_name,
            )
        except AttributeError:
            self.logger.warning(
                f"{error_message} needs to be a string",
                experiment_name=experiment_name,
            )
        except JSONDecodeError:
            self.logger.warning(
                f"{error_message} needs to be a string of a list",
                experiment_name=experiment_name,
            )

        # if feature modules is none, see if we can retrieve them from the old model
        if not model_specs.feature_modules:
            try:
                if loaded_model.feature_modules:
                    model_specs.feature_modules = loaded_model.feature_modules
                    self.logger.info(
                        "feature_modules retrieved from old model with an attribute",
                        experiment_name=experiment_name,
                    )
            except AttributeError:
                self.logger.warning(
                    "feature_modules not an attribute of the old model, using None ",
                    experiment_name=experiment_name,
                )
        return model_specs.feature_modules

    def _get_model_uri(self, artifact_uri: str) -> str:
        """Set model uri based on latest run.

        Note: this function helps to mock during unit tests

        """
        return os.path.join(artifact_uri, "model/")



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from typing import Union

from openstef.enums import MLModelType
from openstef.model.objective import (
    LGBRegressorObjective,
    LinearRegressorObjective,
    ProLoafRegressorObjective,
    RegressorObjective,
    XGBQuantileRegressorObjective,
    XGBRegressorObjective,
    ARIMARegressorObjective,
)
from openstef.model.regressors.custom_regressor import (
    create_custom_objective,
    is_custom_type,
)


class ObjectiveCreator:
    OBJECTIVES = {
        MLModelType.XGB: XGBRegressorObjective,
        MLModelType.LGB: LGBRegressorObjective,
        MLModelType.XGB_QUANTILE: XGBQuantileRegressorObjective,
        MLModelType.ProLoaf: ProLoafRegressorObjective,
        MLModelType.LINEAR: LinearRegressorObjective,
        MLModelType.ARIMA: ARIMARegressorObjective,
    }

    @staticmethod
    def create_objective(model_type: Union[MLModelType, str]) -> RegressorObjective:
        """Create an objective function based on model type.

        Args:
            model_type: Model type to construct.

        Raises:
            NotImplementedError: When using an invalid model_type.

        Returns:
            Objective function

        """
        try:
            # This will raise a ValueError when an invalid model_type str is used
            # and nothing when a MLModelType enum is used.
            if is_custom_type(model_type):
                objective = create_custom_objective(model_type)
            else:
                model_type = MLModelType(model_type)
                objective = ObjectiveCreator.OBJECTIVES[model_type]
        except ValueError as e:
            valid_types = [t.value for t in MLModelType]
            raise NotImplementedError(
                f"No objective for '{model_type}', "
                f"valid model_types are: {valid_types}"
                "or import a custom model"
            ) from e

        return objective



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from typing import Union

import structlog

from openstef.enums import MLModelType
from openstef.model.regressors.custom_regressor import is_custom_type, load_custom_model
from openstef.model.regressors.lgbm import LGBMOpenstfRegressor
from openstef.model.regressors.linear import LinearOpenstfRegressor
from openstef.model.regressors.regressor import OpenstfRegressor
from openstef.model.regressors.xgb import XGBOpenstfRegressor
from openstef.model.regressors.xgb_quantile import XGBQuantileOpenstfRegressor
from openstef.model.regressors.arima import ARIMAOpenstfRegressor

logger = structlog.get_logger(__name__)
try:
    from openstef.model.regressors.proloaf import OpenstfProloafRegressor
except ImportError:
    logger.info("Proloaf not available, setting constructor to None")
    OpenstfProloafRegressor = None

valid_model_kwargs = {
    MLModelType.XGB: [
        "n_estimators",
        "objective",
        "max_depth",
        "learning_rate",
        "verbosity",
        "booster",
        "tree_method",
        "gamma",
        "min_child_weight",
        "max_delta_step",
        "subsample",
        "colsample_bytree",
        "colsample_bylevel",
        "colsample_bynode",
        "reg_alpha",
        "reg_lambda",
        "scale_pos_weight",
        "base_score",
        "missing",
        "num_parallel_tree",
        "kwargs",
        "random_state",
        "n_jobs",
        "monotone_constraints",
        "interaction_constraints",
        "importance_type",
        "gpu_id",
        "validate_parameters",
        "early_stopping_rounds",
    ],
    MLModelType.LGB: [
        "boosting_type",
        "objective",
        "num_leaves",
        "max_depth",
        "learning_rate",
        "n_estimators",
        "subsample_for_bin",
        "min_split_gain",
        "min_child_weight",
        "min_child_samples",
        "subsample",
        "subsample_freq",
        "colsample_bytree",
        "reg_alpha",
        "reg_lambda",
        "random_state",
        "n_jobs",
        "silent",
        "importance_type",
        "early_stopping_rounds",
    ],
    MLModelType.XGB_QUANTILE: [
        "quantiles",
        "gamma",
        "colsample_bytree",
        "subsample",
        "min_child_weight",
        "max_depth",
        "early_stopping_rounds",
    ],
    MLModelType.ProLoaf: [
        "relu_leak",
        "encoder_features",
        "decoder_features",
        "core_layers",
        "rel_linear_hidden_size",
        "rel_core_hidden_size",
        "dropout_fc",
        "dropout_core",
        "training_metric",
        "metric_options",
        "optimizer_name",
        "early_stopping_patience",
        "early_stopping_margin",
        "learning_rate",
        "max_epochs",
        "device",
        "batch_size",
        "history_horizon",
        "horizon_minutes",
    ],
    MLModelType.LINEAR: [
        "missing_values",
        "imputation_strategy",
        "fill_value",
    ],
    MLModelType.ARIMA: [
        "backtest_max_horizon",
        "order",
        "seasonal_order",
        "trend",
    ],
}


class ModelCreator:
    """Factory object for creating machine learning models."""

    # Set object mapping
    MODEL_CONSTRUCTORS = {
        MLModelType.XGB: XGBOpenstfRegressor,
        MLModelType.LGB: LGBMOpenstfRegressor,
        MLModelType.XGB_QUANTILE: XGBQuantileOpenstfRegressor,
        MLModelType.ProLoaf: OpenstfProloafRegressor,
        MLModelType.LINEAR: LinearOpenstfRegressor,
        MLModelType.ARIMA: ARIMAOpenstfRegressor,
    }

    @staticmethod
    def create_model(model_type: Union[MLModelType, str], **kwargs) -> OpenstfRegressor:
        """Create a machine learning model based on model type.

        Args:
            model_type: Model type to construct.
            kwargs: Optional keyword argument to pass to the model.

        Raises:
            NotImplementedError: When using an invalid model_type.

        Returns:
            OpenSTEF model

        """
        try:
            # This will raise a ValueError when an invalid model_type str is used
            # and nothing when a MLModelType enum is used.
            if is_custom_type(model_type):
                model_class = load_custom_model(model_type)
                valid_kwargs = model_class.valid_kwargs()
            else:
                model_type = MLModelType(model_type)
                model_class = ModelCreator.MODEL_CONSTRUCTORS[model_type]
                valid_kwargs = valid_model_kwargs[model_type]
                # Check if model as imported
                if model_class is None:
                    raise ImportError(
                        f"Constructor not available for '{model_type}'. "
                        "Perhaps you forgot to install an optional dependency? "
                        "Please refer to the ReadMe for instructions"
                    )
        except ValueError as e:
            valid_types = [t.value for t in MLModelType]
            raise NotImplementedError(
                f"No constructor for '{model_type}', "
                f"valid model_types are: {valid_types} "
                "or import a custom model"
            ) from e

        # only pass relevant arguments to model constructor to prevent warnings
        model_kwargs = {
            key: value for key, value in kwargs.items() if key in valid_kwargs
        }

        return model_class(**model_kwargs)



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from collections import OrderedDict
from time import perf_counter


class PerformanceMeter:
    def __init__(self, logger):
        self.logger = logger
        self.levels = OrderedDict()
        self.level_timers = []
        self.checkpoint_timers = []

    def start_level(self, level_label: str, level_name: str, **kwargs):
        """Enters a new level in the performance meter and logs it.

        This function also creates a checkpoint on the newly created level.

        Args:
            level_label: The label of the new level. This could i.e. be 'task'
            level_name: The name of the specified level. This could i.e. be
                'tracy_todo'
            **kwargs: Any other kwargs are appended to the logging.

        Returns:
            self

        """
        self.levels[level_label] = level_name

        self.logger.info(
            f"{level_label.capitalize()} started",
            **self.levels,
            **kwargs,
        )
        time = perf_counter()
        self.level_timers.append(time)
        self.checkpoint_timers.append(time)

        return self

    def checkpoint(self, name_checkpoint: str, **kwargs):
        """Creates a timing checkpoint and logs the runtime from the previous one.

        Args:
            name_checkpoint: The name of the checkpoint. This will be logged as
                "checkpoint: name_checkpoint"
            **kwargs: Any other kwargs are appended to the logging.

        Returns:
            self

        """
        runtime = round(perf_counter() - self.checkpoint_timers.pop(), ndigits=3)
        self.logger.info(
            f"{name_checkpoint.capitalize()} completed",
            **self.levels,
            ktp_checkpoint=name_checkpoint,
            ktp_runtime=runtime,
            **kwargs,
        )
        self.checkpoint_timers.append(perf_counter())

        return self

    def complete_level(self, successful: bool = True, **kwargs):
        """Completes the most inner level and logs the total runtime of that level.

        Args:
            successful: Whether the level was successful. Defaults to True.
            **kwargs: Any other kwargs are appended to the logging.

        Returns:
            self

        """
        runtime = round(perf_counter() - self.level_timers.pop(), ndigits=3)
        self.checkpoint_timers.pop()

        level_label, level_name = self.levels.popitem()

        self.logger.info(
            f"{level_label.capitalize()} completed",
            **self.levels,
            **{level_label: level_name},
            ktp_runtime=runtime,
            ktp_successful=int(successful),
            **kwargs,
        )

        return self



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from datetime import datetime

import numpy as np
import pandas as pd
import structlog
from scipy import stats
from sklearn.base import RegressorMixin

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.exceptions import ModelWithoutStDev


class ConfidenceIntervalApplicator:
    def __init__(self, model: RegressorMixin, forecast_input_data: pd.DataFrame):
        self.model = model
        self.forecast_input_data = forecast_input_data
        self.logger = structlog.get_logger(self.__class__.__name__)

    def add_confidence_interval(
        self,
        forecast: pd.DataFrame,
        pj: PredictionJobDataClass,
    ) -> pd.DataFrame:
        """Add a confidence interval to a forecast.

        Adds a confidence interval to a forecast in two ways:
            1. "stdev" column, this is a column with a standard deviation that is
                determined during training (ConfidenceGenerator)
            2. Quantile columns, these columns give a more precise defenition of the
                confidence interval. Quantile columns are determined with one of two
                methods, depending on the model type group:

                a. Default, using the "stdev" column and the assumption the error is
                    normally distributed.
                b. Quantile regression, this method is only available for quantile
                    models and uses specifically trained models to estimate the
                    quantiles of the confidence interval.

                Depending on the model type (quantile or non quantile),
                 a confidence interval is added to the forecast based on quantile
                 regression or the default method.

        Args:
            forecast: Forecast DataFrame with columns: "forecast"
            pj: Prediction job

        Returns:
            Forecast DataFrame with columns; "forecast", "stdev" and quantile columns.

        """
        temp_forecast = self._add_standard_deviation_to_forecast(forecast)

        if self.model.can_predict_quantiles:
            return self._add_quantiles_to_forecast_quantile_regression(
                temp_forecast, pj["quantiles"]
            )

        return self._add_quantiles_to_forecast_default(temp_forecast, pj["quantiles"])

    def _add_standard_deviation_to_forecast(
        self, forecast: pd.DataFrame
    ) -> pd.DataFrame:
        """Add a standard deviation to a live forecast.

        The stdev for intermediate forecast horizons is interpolated.

        For the input standard_deviation, it is preferred that the forecast horizon is
        expressed in Hours, instead of 'Near/Far'. For now, Near/Far is supported,
        but this will be deprecated.

        Args:
            forecast: Forecast DataFrame with columns: "forecast"

        Returns:
            Forecast with added standard deviation. DataFrame with columns:
                "forecast", "stdev"

        """
        minimal_resolution: int = 15  # Minimal time resolution in minutes
        standard_deviation = self.model.standard_deviation

        # raise an exception if no valid standard deviation is available
        if standard_deviation is None:
            raise ModelWithoutStDev("No stdev available")

        if standard_deviation.empty:  # make separate statement to avoid None.empty
            raise ModelWithoutStDev("No stdev available")

        if standard_deviation.stdev.isnull().values.all():
            raise ModelWithoutStDev("All stdev values are NA")

        # Fill stdev nans with the mean of all stdev values
        if standard_deviation.stdev.isnull().values.any():
            self.logger.warning(
                "Stdev for some hours is not known, filling in with mean."
            )
            standard_deviation["stdev"] = standard_deviation.stdev.fillna(
                standard_deviation.stdev.mean()
            )

        # pivot to have a dataframe with columns [stdev, hour, horizon] for a
        # 'near' and a 'far' horizon
        stdev = standard_deviation.pivot_table(columns=["horizon"], index="hour")[
            "stdev"
        ]
        # Prepare input dataframes for near and far horizon
        near = stdev.columns.min()
        far = stdev.columns.max()

        forecast_copy = forecast.copy()
        # add time ahead column if not already present
        if "tAhead" not in forecast_copy.columns:
            # Determine now, rounded on 15 minutes,
            # Rounding helps to prevent fractional t_aheads
            now = (
                pd.Series(datetime.utcnow().replace(tzinfo=forecast_copy.index.tzinfo))
                .min()
                .round(f"{minimal_resolution}T")
                .to_pydatetime()
            )
            # Determine t_aheads by subtracting with now
            forecast_copy["tAhead"] = (
                forecast_copy.index - now
            ).total_seconds() / 3600.0

        # add helper column hour
        forecast_copy["hour"] = forecast_copy.index.hour

        # Define functions which can be used to approximate the error for in-between
        # time horizons
        # Let's fit and exponential decay of accuracy
        def calc_exp_dec(t, stdev_row, near, far):
            # We use the formula sigma(t) = (1 - A * exp(-t/tau)) + b
            # Strictly speaking, tau is specific for each time series.
            # However, for simplicity, we use tau = Far/4.
            # This represents a situation where the stdev at 25% of the Far horizon,
            # has increased by two.
            tau = far / 4.0
            # Filling in the known sigma(Near) and sigma(Far) gives:
            sf, sn = stdev_row[far], stdev_row[near]
            A = (sf - sn) / ((1 - np.exp(-far / tau)) - (1 - np.exp(-near / tau)))
            b = sn - A * (1 - np.exp(-near / tau))
            return A * (1 - np.exp(-t / tau)) + b

        # If only one horizon is available use that one
        if len(stdev.columns) == 1:
            forecast_copy["stdev"] = forecast_copy.apply(
                lambda x: stdev.loc[x.hour], axis=1
            )
        # If more are available do something fancy with interpolation
        else:
            # Add stdev to forecast_copy dataframe
            forecast_copy["stdev"] = forecast_copy.apply(
                lambda x: calc_exp_dec(x.tAhead, stdev.loc[x.hour], near, far), axis=1
            )
        return forecast_copy.drop(columns=["hour"])

    @staticmethod
    def _add_quantiles_to_forecast_default(
        forecast: pd.DataFrame, quantiles: list[float]
    ) -> pd.DataFrame:
        """Add quantiles to forecast.

            Use the standard deviation to calculate the quantiles.

        Args:
            forecast: Forecast (should contain a 'forecast' + 'stdev' column)
            quantiles: List with desired quantiles,
                for example: [0.01, 0.1, 0.9, 0.99]

        Returns:
            Forecast DataFrame with quantile (e.g. 'quantile_PXX')
                columns added.

        """
        # Check if stdev and forecast are in the dataframe
        if not all(elem in forecast.columns for elem in ["forecast", "stdev"]):
            raise ValueError("Forecast should contain a 'forecast' and 'stdev' column")

        for quantile in quantiles:
            quantile_key = f"quantile_P{quantile * 100:02.0f}"
            forecast[quantile_key] = (
                forecast["forecast"] + stats.norm.ppf(quantile) * forecast["stdev"]
            )

        return forecast

    def _add_quantiles_to_forecast_quantile_regression(
        self, forecast: pd.DataFrame, quantiles: list[float]
    ) -> pd.DataFrame:
        """Add quantiles to forecast.

            Use trained quantile regression model to calculate the quantiles.

        Args:
            forecast: Forecast
            quantiles: List with desired quantiles

        Returns:
            Forecast DataFrame with quantile (e.g. 'quantile_PXX')
                columns added.

        """
        # Only determine quantiles for datetimes in forecast
        quantile_df = pd.DataFrame(index=self.forecast_input_data.index)
        for quantile in quantiles:
            quantile_key = f"quantile_P{quantile * 100:02.0f}"
            quantile_df[quantile_key] = self.model.predict(
                self.forecast_input_data, quantile=quantile
            )

        return forecast.merge(
            quantile_df, left_index=True, right_index=True, how="left"
        )



# SPDX-FileCopyrightText: 2017-2023 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from typing import Any, Union

import numpy as np
import pandas as pd

# These imports will require the proloaf optional dependencies to be installed
import proloaf.datahandler as dh
import torch
from proloaf.modelhandler import ModelWrapper

from openstef.model.regressors.regressor import OpenstfRegressor

# TODO: implement the hyperparameter optimalisation via optuna
# TODO: set the default for hyperparameters in the init of OpenstfProloafRegressor
# TODO: implement function for defining encoder and decoder features


def divide_scaling_groups(x: pd.DataFrame) -> tuple[list[str], list[str], list[str]]:
    """Divides the column names over different type of scaling groups.

    Args:
        x: Dataframe from which columns have to be divided

    Returns:
        List of all the grouped features for scaling (three groups)

    """
    minmax_scale_features = []
    oh_scale_features = []
    no_scale_features = []

    for column in x.columns:
        if (x[column].min() <= -1) or (x[column].max() >= 1):
            minmax_scale_features.append(column)
        elif x[column].dtype == "bool":
            oh_scale_features.append(column)
        else:
            no_scale_features.append(column)

    return minmax_scale_features, oh_scale_features, no_scale_features


def apply_scaling(
    scaler_features: tuple[list[str], list[str], list[str]],
    data: pd.DataFrame,
    scalers=None,
):
    """Applies different scaling methods to a certain dataframe (minmax, one hot, or no scaling).

    Args:
        scaler_features: Three different lists with features for each scaling
        x: Dataframe from which columns have to be divided
        scalers: scalers resulting from the previous scaling

    Returns:
        Dataframe with all the scaled features

    """
    selected_features, scalers = dh.scale_all(
        data,
        scalers=scalers,
        feature_groups=[
            {
                "name": "main",
                "scaler": ["minmax", -1.0, 1.0],
                "features": scaler_features[0],
            },
            {"name": "aux", "scaler": None, "features": scaler_features[2]},
        ],
    )

    # One hot encoding certain features
    onehot_feature_groups = [
        {
            "name": "main",
            "scaler": [
                "onehot",
            ],
            "features": scaler_features[1],
        }
    ]
    for group in onehot_feature_groups:
        df_onehot = data.filter(group["features"])
        result_oh_scale = np.transpose(np.array(df_onehot.iloc[:, :], dtype=np.int))
        df_onehot.iloc[:, :] = result_oh_scale.T

    if not df_onehot.columns.empty:
        selected_features = pd.concat([selected_features, df_onehot], axis=1)
    data = selected_features.iloc[:, :].replace(np.nan, 0)

    return data, scalers


class OpenstfProloafRegressor(OpenstfRegressor, ModelWrapper):
    def __init__(
        self,
        name: str = "model",
        core_net: str = "torch.nn.LSTM",
        relu_leak: float = 0.1,
        encoder_features: list[str] = [
            "historic_load",
        ],  # make sure historic load is present, TODO: implement so you can use None
        decoder_features: list[str] = [
            "air_density"
        ],  # TODO: implement so you can use None
        core_layers: int = 1,
        rel_linear_hidden_size: float = 1.0,
        rel_core_hidden_size: float = 1.0,
        dropout_fc: float = 0.4,
        dropout_core: float = 0.3,
        training_metric: str = "nllgauss",
        metric_options: dict[str, Any] = {},
        optimizer_name: str = "adam",
        early_stopping_patience: int = 7,
        early_stopping_margin: float = 0,
        learning_rate: float = 1e-3,
        max_epochs: int = 100,
        device: Union[str, int] = "cpu",  # "cuda" or "cpu"
        batch_size: int = 6,
        history_horizon: int = 24,
        horizon_minutes: int = 2880,  # 2 days in minutes,
    ):
        self.device = device
        self.batch_size = batch_size
        self.history_horizon = history_horizon
        self.forecast_horizon = int(horizon_minutes / 60)
        ModelWrapper.__init__(
            self,
            name=name,
            core_net=core_net,
            relu_leak=relu_leak,
            encoder_features=encoder_features,
            decoder_features=decoder_features,
            core_layers=core_layers,
            rel_linear_hidden_size=rel_linear_hidden_size,
            rel_core_hidden_size=rel_core_hidden_size,
            dropout_fc=dropout_fc,
            dropout_core=dropout_core,
            training_metric=training_metric,
            metric_options=metric_options,
            optimizer_name=optimizer_name,
            early_stopping_patience=early_stopping_patience,
            early_stopping_margin=early_stopping_margin,
            learning_rate=learning_rate,
            max_epochs=max_epochs,
        )
        self.to(device)

    @property
    def feature_names(self):
        return (
            ["load"] + self.encoder_features + self.decoder_features
        )  # TODO: gehele range, of een enkele feature

    @property
    def can_predict_quantiles(self):
        return False

    def predict(self, x: pd.DataFrame) -> np.ndarray:
        x = x[list(self.feature_names)[1:]]
        # Apply scaling and interpolation for NaN values
        x = dh.fill_if_missing(x, periodicity=24)
        x, _ = apply_scaling(
            [
                self.minmax_scale_features,
                self.oh_scale_features,
                self.no_scale_features,
            ],
            x,
            self.scalers,
        )

        inputs_enc = torch.tensor(
            x[self.encoder_features].to_numpy(), dtype=torch.float
        ).unsqueeze(dim=0)
        inputs_dec = torch.tensor(
            x[self.decoder_features].to_numpy(), dtype=torch.float
        ).unsqueeze(dim=0)
        prediction = (
            ModelWrapper.predict(self, inputs_enc, inputs_dec)[:, :, 0]
            .squeeze()
            .detach()
            .numpy()
        )

        return prediction

    def fit(
        self,
        x: pd.DataFrame,
        y: pd.DataFrame,
        eval_set: tuple = None,
        early_stopping_rounds: int = None,
        verbose: bool = False,
        **kwargs,
    ) -> ModelWrapper:
        x = x[list(self.feature_names)[1:]]
        # Apply scaling and interpolation for NaN values
        x = dh.fill_if_missing(x, periodicity=24)
        (
            self.minmax_scale_features,
            self.oh_scale_features,
            self.no_scale_features,
        ) = divide_scaling_groups(x)
        x, self.scalers = apply_scaling(
            [
                self.minmax_scale_features,
                self.oh_scale_features,
                self.no_scale_features,
            ],
            x,
            self.scalers,
        )
        y = y.to_frame()
        self.target_id = [y.columns[0]]

        df_train = pd.concat([x, y], axis="columns", verify_integrity=True)

        print(f"{self.encoder_features = }")
        print(f"{self.decoder_features = }")

        train_dl, _, _ = dh.transform(
            df=df_train,
            encoder_features=self.encoder_features,
            decoder_features=self.decoder_features,
            batch_size=self.batch_size,
            history_horizon=self.history_horizon,
            forecast_horizon=self.forecast_horizon,
            target_id=self.target_id,
            train_split=1.0,
            validation_split=1.0,
            device=self.device,
        )

        val_x, val_y = eval_set[1][0], eval_set[1][1]
        val_x = dh.fill_if_missing(val_x, periodicity=24)
        val_x, _ = apply_scaling(
            [
                self.minmax_scale_features,
                self.oh_scale_features,
                self.no_scale_features,
            ],
            val_x,
            self.scalers,
        )

        df_val = pd.concat([val_x, val_y], axis="columns", verify_integrity=True)
        _, validation_dl, _ = dh.transform(
            df=df_val,
            encoder_features=self.encoder_features,
            decoder_features=self.decoder_features,
            batch_size=self.batch_size,
            history_horizon=self.history_horizon,
            forecast_horizon=self.forecast_horizon,
            target_id=self.target_id,
            train_split=0.0,
            validation_split=1.0,
            device=self.device,
        )
        self.to(self.device)
        self.init_model()
        self.is_fitted_ = True

        return self.run_training(train_dl, validation_dl)

    def get_params(self, deep=True):
        model_params = self.get_model_config()
        training_params = self.get_training_config()
        return {**model_params, **training_params}

    def set_params(self, **params):
        self.update(**params)
        return self


if __name__ == "__main__":
    test = OpenstfProloafRegressor()
    print(test)



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the CRON job that is periodically executed to make prognoses and save them in to the database.

This code assumes trained models are available from the persistent storage. If these
are not available run model_train.py to train all models.
To provide the prognoses the folowing steps are carried out:
  1. Get historic training data (TDCV, Load, Weather and APX price data)
  2. Apply features
  3. Load model
  4. Make prediction
  5. Write prediction to the database
  6. Send Teams message if something goes wrong

Example:
    This module is meant to be called directly from a CRON job.
    Alternatively this code can be run directly by running::

        $ python create_forecast.py

"""
from datetime import datetime, timedelta
from pathlib import Path

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.enums import MLModelType, PipelineType
from openstef.pipeline.create_forecast import create_forecast_pipeline
from openstef.tasks.utils.predictionjobloop import PredictionJobLoop
from openstef.tasks.utils.taskcontext import TaskContext

T_BEHIND_DAYS: int = 14
T_AHEAD_DAYS: int = 2


def create_forecast_task(pj: PredictionJobDataClass, context: TaskContext) -> None:
    """Top level task that creates a forecast.

    On this task level all database and context manager dependencies are resolved.

    Expected prediction job keys; "id", "lat", "lon", "resolution_minutes",
        "horizon_minutes", "type", "name", "quantiles"

    Args:
        pj: Prediction job
        context: Contect object that holds a config manager and a database connection

    """
    # Check pipeline types
    if PipelineType.FORECAST not in pj.pipelines_to_run:
        context.logger.info(
            "Skip this PredictionJob because forecast pipeline is not specified in the pj."
        )
        return

    # Extract mlflow tracking URI and trained models folder
    mlflow_tracking_uri = context.config.paths_mlflow_tracking_uri

    # Define datetime range for input data
    datetime_start = datetime.utcnow() - timedelta(days=T_BEHIND_DAYS)
    datetime_end = datetime.utcnow() + timedelta(days=T_AHEAD_DAYS)

    # Retrieve input data
    input_data = context.database.get_model_input(
        pid=pj["id"],
        location=[pj["lat"], pj["lon"]],
        datetime_start=datetime_start,
        datetime_end=datetime_end,
    )
    # Make forecast with the forecast pipeline
    forecast = create_forecast_pipeline(
        pj, input_data, mlflow_tracking_uri=mlflow_tracking_uri
    )

    # Write forecast to the database
    context.database.write_forecast(forecast, t_ahead_series=True)


def main(model_type=None, config=None, database=None):
    taskname = Path(__file__).name.replace(".py", "")

    if database is None or config is None:
        raise RuntimeError(
            "Please specify a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    with TaskContext(taskname, config, database) as context:

        if model_type is None:
            model_type = [ml.value for ml in MLModelType]

        PredictionJobLoop(context, model_type=model_type).map(
            create_forecast_task, context
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""This module contains the CRON job that is periodically executed to make prognoses of wind features.

These features are usefull for splitting the load in solar and wind contributions and
making prognoses.

Example:
    This module is meant to be called directly from a CRON job. A description of the
    CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::
        $ python create_wind_forecast

"""
from pathlib import Path

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.feature_engineering import weather_features
from openstef.tasks.utils.predictionjobloop import PredictionJobLoop
from openstef.tasks.utils.taskcontext import TaskContext


def make_wind_forecast_pj(pj: PredictionJobDataClass, context: TaskContext) -> None:
    """Make a wind prediction for a specific prediction job.

    Args:
        pj: Prediction job
        context: Context manager

    """
    context.logger.info("Get turbine data", turbine_type=pj["turbine_type"])
    turbine_data = context.database.get_power_curve(pj["turbine_type"])

    context.logger.info(
        "Get windspeed", location=[pj["lat"], pj["lon"]], hub_height=pj["hub_height"]
    )
    windspeed = context.database.get_wind_input(
        (pj["lat"], pj["lon"]),
        pj["hub_height"],
        pj["horizon_minutes"],
        pj["resolution_minutes"],
    )

    context.logger.info("Calculate windturbine power", n_turbines=pj["n_turbines"])
    power = weather_features.calculate_windturbine_power_output(
        windspeed, pj["n_turbines"], turbine_data
    ).rename(columns=dict(windspeed_100m="forecast"))

    context.logger.info("Store wind prediction in database")
    power["pid"] = pj["id"]
    power["type"] = "wind"
    power["algtype"] = "powerCurve"
    power["customer"] = pj["name"]
    power["description"] = pj["description"]
    context.database.write_forecast(power, t_ahead_series=True)


def main(config=None, database=None):
    taskname = Path(__file__).name.replace(".py", "")

    if database is None or config is None:
        raise RuntimeError(
            "Please specifiy a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    with TaskContext(taskname, config, database) as context:
        context.logger.info("Querying wind prediction jobs from database")
        prediction_jobs = context.database.get_prediction_jobs_wind()
        prediction_jobs = [x for x in prediction_jobs if x["model"] == "latest"]

        PredictionJobLoop(context, prediction_jobs=prediction_jobs).map(
            make_wind_forecast_pj, context
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""This module contains the CRON job that is periodically executed to calculate key performance indicators (KPIs).

This code assumes prognoses are available from the persistent storage.
If these are not available run create_forecast.py to train all models.

The folowing tasks are caried out:
  1: Calculate the KPI for a given pid. Ignore SplitEnergy
  2: Create figures
  3: Write KPI to database

Example:
    This module is meant to be called directly from a CRON job.
    Alternatively this code can be run directly by running::
        $ python calculate_kpi.py

"""
# Import builtins
from datetime import datetime, timedelta
from pathlib import Path

import numpy as np
import pandas as pd
import structlog

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.enums import MLModelType
from openstef.exceptions import NoPredictedLoadError, NoRealisedLoadError
from openstef.metrics import metrics
from openstef.tasks.utils.predictionjobloop import PredictionJobLoop
from openstef.tasks.utils.taskcontext import TaskContext
from openstef.validation import validation

# Thresholds for retraining and optimizing
THRESHOLD_RETRAINING = 0.25
THRESHOLD_OPTIMIZING = 0.50


def main(model_type: MLModelType = None, config=None, database=None) -> None:
    taskname = Path(__file__).name.replace(".py", "")

    if database is None or config is None:
        raise RuntimeError(
            "Please specifiy a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    if model_type is None:
        model_type = [ml.value for ml in MLModelType]

    with TaskContext(taskname, config, database) as context:
        # Set start and end time
        start_time = datetime.utcnow() - timedelta(days=1)
        end_time = datetime.utcnow()

        PredictionJobLoop(context, model_type=model_type).map(
            check_kpi_task,
            context,
            start_time=start_time,
            end_time=end_time,
        )


def check_kpi_task(
    pj: PredictionJobDataClass,
    context: TaskContext,
    start_time: datetime,
    end_time: datetime,
) -> None:
    # Apply default parameters if none are provided
    if start_time is None:
        start_time = datetime.utcnow() - timedelta(days=1)
    if end_time is None:
        end_time = datetime.utcnow()

    # Get realised load data
    realised = context.database.get_load_pid(pj["id"], start_time, end_time, "15T")

    # Get predicted load
    predicted_load = context.database.get_predicted_load_tahead(
        pj, start_time, end_time
    )

    # Get basecase prediction
    load_1_week_before = context.database.get_load_pid(
        pj["id"], start_time - timedelta(days=7), end_time - timedelta(days=7), "15T"
    )
    if len(load_1_week_before) > 0:
        basecase = load_1_week_before.shift(periods=7, freq="d")
    else:
        basecase = pd.DataFrame()

    kpis = calc_kpi_for_specific_pid(pj["id"], realised, predicted_load, basecase)
    # Write KPI's to database
    context.database.write_kpi(pj, kpis)

    # Add pid to the list of pids that should be retrained or optimized if
    # performance is insufficient
    if kpis["47.0h"]["rMAE"] > THRESHOLD_RETRAINING:
        context.logger.warning(
            "Need to retrain model, retraining threshold rMAE 47h exceeded",
            t_ahead="47.0h",
            rMAE=kpis["47.0h"]["rMAE"],
            retraining_threshold=THRESHOLD_RETRAINING,
        )
        function_name = "train_model"

        context.logger.info("Adding tracy job", function=function_name)
        context.database.ktp_api.add_tracy_job(pj["id"], function=function_name)

    if kpis["47.0h"]["rMAE"] > THRESHOLD_OPTIMIZING:
        context.logger.warning(
            "Need to optimize hyperparameters, optimizing threshold rMAE 47h exceeded",
            t_ahead="47.0h",
            rMAE=kpis["47.0h"]["rMAE"],
            optimizing_threshold=THRESHOLD_OPTIMIZING,
        )
        function_name = "optimize_hyperparameters"
        context.logger.info("Adding tracy job", function=function_name)
        context.database.ktp_api.add_tracy_job(pj["id"], function=function_name)


def calc_kpi_for_specific_pid(
    pid: int,
    realised: pd.DataFrame,
    predicted_load: pd.DataFrame,
    basecase: pd.DataFrame,
) -> dict:
    """Function that checks the model performance based on a pid. This function.

    - loads and combines forecast and realised data
    - calculated several key performance indicators (KPIs)
    These metric include:
        - RMSE,
        - bias,
        - NSME (model efficiency, between -inf and 1)
        - Mean absolute Error

    Args:
        pid: Prediction ID for a given prediction job
        realised: Realised load.
        predicted_load: Predicted load.
        basecase: Basecase predicted load.

    Returns:
        - Dictionary that includes a dictonary for each t_ahead.
        - Dict includes enddate en window (in days) for clarification

    Raises:
        NoPredictedLoadError: When no predicted load for given datatime range.
        NoRealisedLoadError: When no realised load for given datetime range.

    Example:
        To get the rMAE for the 24 hours ahead prediction: kpis['24h']['rMAE']

    """
    COMPLETENESS_REALISED_THRESHOLDS = 0.7
    COMPLETENESS_PREDICTED_LOAD_THRESHOLD = 0.7

    log = structlog.get_logger(__name__)

    # If predicted is empty
    if len(predicted_load) == 0:
        raise NoPredictedLoadError(pid)

    # If realised is empty
    if len(realised) == 0:
        raise NoRealisedLoadError(pid)

    # Define start and end time
    start_time = realised.index.min().to_pydatetime()
    end_time = realised.index.max().to_pydatetime()

    completeness_realised = validation.calc_completeness_dataframe(realised)[0]

    # Interpolate missing data if needed
    realised = realised.resample("15T").interpolate(limit=3)

    completeness_predicted_load = validation.calc_completeness_dataframe(predicted_load)

    # Combine the forecast and the realised to make sure indices are matched nicely
    combined = pd.merge(realised, predicted_load, left_index=True, right_index=True)

    # Add basecase (load in same time period 7 days ago)
    # Check if basecase is not empty, else make a dummy dataframe
    if len(basecase) == 0:
        basecase = pd.DataFrame(columns=["load"])
    basecase = basecase.rename(columns=dict(load="basecase"))

    combined = combined.merge(basecase, how="left", left_index=True, right_index=True)

    # Raise exception in case of constant load
    if combined.load.nunique() == 1:
        structlog.get_logger(__name__).warning(
            "The load is constant! KPIs will still be calculated, but relative metrics"
            " will be nan"
        )

    # Define output dictonary
    kpis = dict()

    # Extract t_aheads from predicted_load,
    # Make a list of tuples with [(forecast_xh, stdev_xh),(..,..),..]
    hor_list = [
        ("forecast_" + t_ahead, "stdev_" + t_ahead)
        for t_ahead in set(col.split("_")[1] for col in predicted_load.columns)
    ]

    # cast date to int
    date = pd.to_datetime(end_time)

    # Calculate model metrics and add them to the output dictionary
    log.info("Start calculating kpis")
    for hor_cols in hor_list:
        t_ahead_h = hor_cols[0].split("_")[1]
        fc = combined[hor_cols[0]]  # load predictions
        st = combined[hor_cols[1]]  # standard deviations of load predictions

        completeness_predicted_load_specific_hor = (
            validation.calc_completeness_dataframe(fc.to_frame(name=t_ahead_h))[0]
        )
        kpis.update(
            {
                t_ahead_h: {
                    "RMSE": metrics.rmse(combined["load"], fc),
                    "bias": metrics.bias(combined["load"], fc),
                    "NSME": metrics.nsme(combined["load"], fc),
                    "MAE": metrics.mae(combined["load"], fc),
                    "rMAE": metrics.r_mae(combined["load"], fc),
                    "rMAE_highest": metrics.r_mae_highest(combined["load"], fc),
                    "rMNE_highest": metrics.r_mne_highest(combined["load"], fc),
                    "rMPE_highest": metrics.r_mpe_highest(combined["load"], fc),
                    "rMAE_lowest": metrics.r_mae_lowest(combined["load"], fc),
                    "skill_score_basecase": metrics.skill_score(
                        combined["load"],
                        combined["basecase"],
                        np.mean(combined["basecase"]),
                    ),
                    "skill_score": metrics.skill_score(
                        combined["load"], fc, np.mean(combined["basecase"])
                    ),
                    "skill_score_positive_peaks": metrics.skill_score_positive_peaks(
                        combined["load"], fc, np.mean(combined["basecase"])
                    ),
                    "skill_score_positive_peaks_basecase": metrics.skill_score_positive_peaks(
                        combined["load"],
                        combined["basecase"],
                        np.mean(combined["basecase"]),
                    ),
                    "franks_skill_score": metrics.franks_skill_score(
                        combined["load"], fc, combined["basecase"]
                    ),
                    "franks_skill_score_peaks": metrics.franks_skill_score_peaks(
                        combined["load"], fc, combined["basecase"]
                    ),
                    "load_range": combined["load"].max() - combined["load"].min(),
                    "frac_in_1sdev": metrics.frac_in_stdev(combined["load"], fc, st),
                    "frac_in_2sdev": metrics.frac_in_stdev(
                        combined["load"], fc, 2 * st
                    ),
                    "completeness_realised": completeness_realised,
                    "completeness_predicted": completeness_predicted_load_specific_hor,
                    "date": date,  # cast to date
                    "window_days": np.round(
                        (end_time - start_time).total_seconds() / 60.0 / 60.0 / 24.0
                    ),
                }
            }
        )

        if completeness_realised < COMPLETENESS_REALISED_THRESHOLDS:
            log.warning(
                "Completeness realised load too low",
                prediction_id=pid,
                start_time=start_time,
                end_time=end_time,
                completeness=completeness_realised,
                completeness_threshold=COMPLETENESS_REALISED_THRESHOLDS,
            )
            set_incomplete_kpi_to_nan(kpis, t_ahead_h)
        if completeness_predicted_load.any() < COMPLETENESS_PREDICTED_LOAD_THRESHOLD:
            log.warning(
                "Completeness predicted load of specific horizon too low",
                prediction_id=pid,
                horizon=t_ahead_h,
                start_time=start_time,
                end_time=end_time,
                completeness=completeness_predicted_load,
                completeness_threshold=COMPLETENESS_PREDICTED_LOAD_THRESHOLD,
            )
            set_incomplete_kpi_to_nan(kpis, t_ahead_h)

    # Return output dictionary
    return kpis


def set_incomplete_kpi_to_nan(kpis: dict, t_ahead_h: str) -> None:
    """Checks the given kpis for completeness and sets to nan if this not true.

    Args:
        kpis: the kpis
        t_ahead_h: t_ahead_h

    """
    kpi_metrics = list(kpis[t_ahead_h].keys())
    # Set to nan
    for kpi in kpi_metrics:
        if kpi not in [
            "completeness_realised",
            "completeness_predicted",
            "date",
            "window_days",
        ]:
            kpis[t_ahead_h].update({kpi: np.nan})


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin

MINIMAL_RESOLUTION: int = 15  # Used for validating the forecast input


class BaseCaseModel(BaseEstimator, RegressorMixin):
    def predict(self, forecast_input_data: pd.DataFrame) -> pd.DataFrame:
        """Predict using the basecase method. The basecase forecast is determined by the T-7d and T-14d load.

        This means fitting the model is not required. However a fit method is still included to be fully comatible with sklearn.

        Args:
            forecast_input_data: Forecast input dataframe

        Returns:
            Basecase forecast

        """
        return self.make_basecase_forecast(forecast_input_data)

    def fit(self):
        return self

    @staticmethod
    def make_basecase_forecast(
        forecast_input_data: pd.DataFrame, overwrite_delay_hours: int = 48
    ) -> pd.DataFrame:
        """Make a basecase forecast.

        The idea of the basecase forecast is that if all else fails, this forecasts is
        still available. Basecase example: the load of last week.

        Args:
            forecast_input_data: Forecast input dataframe
            overwrite_delay_hours: times before this in the future are not
                forecasted

        Raises:
            ValueError: if columns T-7d or T-14d is not present
            ValueError: If the start of the forecast is before the horizon of the regular forecast
        Returns:
            Basecase forecast

        """
        # Check if required features are provided
        if not all(
            item in forecast_input_data.columns.to_list() for item in ["T-14d", "T-7d"]
        ):
            raise ValueError(
                "Could not make basecase, features T-7d and T-14d are required! Tip:"
                " Generate these features with a FeatureApplicator object."
            )

        # Make basecase forecast: Use load of last week
        basecase_forecast = (
            forecast_input_data[["T-7d"]].dropna().rename(columns={"T-7d": "forecast"})
        )

        # Maybe there is still missing data, for example if the cdb has been down for a
        # while in this case, use the load of 2 weeks before
        basecase_forecast = pd.concat(
            [
                basecase_forecast,
                forecast_input_data[["T-14d"]]
                .dropna()
                .rename(columns={"T-14d": "forecast"}),
            ]
        )
        basecase_forecast = basecase_forecast[
            np.invert(basecase_forecast.index.duplicated())
        ]

        return basecase_forecast.sort_index()

    @property
    def can_predict_quantiles(self):
        return False



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import copy
from datetime import datetime
from typing import Any, Callable, Optional

import optuna
import pandas as pd

from openstef.enums import MLModelType
from openstef.metrics import metrics
from openstef.metrics.reporter import Report, Reporter
from openstef.model.regressors.regressor import OpenstfRegressor
from openstef.model.standard_deviation_generator import StandardDeviationGenerator
from openstef.model_selection.model_selection import split_data_train_validation_test

EARLY_STOPPING_ROUNDS: int = 10
TEST_FRACTION: float = 0.15
VALIDATION_FRACTION: float = 0.15
# See https://xgboost.readthedocs.io/en/latest/parameter.html for all possibilities
EVAL_METRIC: str = "mae"

# https://optuna.readthedocs.io/en/stable/faq.html#objective-func-additional-args


class RegressorObjective:
    """Regressor optuna objective function.

    Use any of the derived classes for optimization using an optuna study.
    The constructor is used to set the "input_data", specify the splitting function
    and its arguments and optionally add some configuration.
    Next the instance will be called by he optuna study during optimization.

    Example usage:

    .. code-block:: py

        # initialize a (derived class) objective function
        objective = XGBRegressorObjective(input_data, test_fraction)
        # use the objective function
        study.optimize(objective)

    """

    def __init__(
        self,
        model: OpenstfRegressor,
        input_data: pd.DataFrame,
        split_func: Optional[Callable] = None,
        split_args: Optional[dict[str, Any]] = None,
        test_fraction=TEST_FRACTION,
        validation_fraction=VALIDATION_FRACTION,
        eval_metric=EVAL_METRIC,
        verbose=False,
    ):
        self.input_data = input_data
        self.train_data = None
        self.validation_data = None
        self.test_data = None
        self.model = model
        self.start_time = datetime.utcnow()
        self.test_fraction = test_fraction
        self.validation_fraction = validation_fraction
        self.eval_metric = eval_metric
        self.eval_metric_function = metrics.get_eval_metric_function(eval_metric)
        self.verbose = verbose
        # Should be set on a derived classes
        self.model_type = None
        self.track_trials = {}

        # split function and arguments
        self.split_func = split_func
        self.split_args = split_args

        # default behavior for splitting
        if self.split_func is None:
            self.split_func = split_data_train_validation_test
            self.split_args = None

    def __call__(
        self,
        trial: optuna.trial.FrozenTrial,
    ) -> float:
        """Optuna objective function.

        Args: trial

        Returns:
            Mean absolute error for this trial.

        """
        # Perform data preprocessing
        split_args = self.split_args
        if split_args is None:
            split_args = {
                "stratification_min_max": self.model_type != MLModelType.ProLoaf,
                "back_test": True,
            }
        (self.train_data, self.validation_data, self.test_data,) = self.split_func(
            self.input_data,
            test_fraction=self.test_fraction,
            validation_fraction=self.validation_fraction,
            **split_args,
        )

        # Test if first column is "load" and last column is "horizon"
        if (
            self.train_data.columns[0] != "load"
            or self.train_data.columns[-1] != "horizon"
        ):
            raise RuntimeError(
                "Column order in train input data not as expected, "
                "could not train a model!"
            )

        # Split in x, y data (x are the features, y is the load)
        train_x, train_y = self.train_data.iloc[:, 1:-1], self.train_data.iloc[:, 0]
        valid_x, valid_y = (
            self.validation_data.iloc[:, 1:-1],
            self.validation_data.iloc[:, 0],
        )
        test_x, test_y = self.test_data.iloc[:, 1:-1], self.test_data.iloc[:, 0]

        # Configure evals for early stopping
        eval_set = [(train_x, train_y), (valid_x, valid_y)]

        # get the parameters used in this trial
        hyper_params = self.get_params(trial)

        # insert parameters into model
        self.model.set_params(**hyper_params)

        # create the specific pruning callback
        pruning_callback = self.get_pruning_callback(trial)
        if pruning_callback is None:
            callbacks = None
        else:
            callbacks = [pruning_callback]

        # validation_0 and validation_1 are available
        self.model.fit(
            train_x,
            train_y,
            eval_set=eval_set,
            early_stopping_rounds=EARLY_STOPPING_ROUNDS,
            verbose=self.verbose,
            eval_metric=self.eval_metric,
            callbacks=callbacks,
        )

        self.model.feature_importance_dataframe = self.model.set_feature_importance()

        # Do confidence interval determination
        self.model = StandardDeviationGenerator(
            self.validation_data
        ).generate_standard_deviation_data(self.model)

        forecast_y = self.model.predict(test_x)
        score = self.eval_metric_function(test_y, forecast_y)

        # Convert float32 to float because float32 is not JSON serializable
        self.track_trials[f" trial: {trial.number}"] = {
            "score": float(score),
            "params": hyper_params,
        }
        trial.set_user_attr(key="model", value=copy.deepcopy(self.model))
        return score

    def get_params(self, trial: optuna.trial.FrozenTrial) -> dict:
        """Get parameters for objective without model specific get_params function.

        Args: trial

        Returns:
            Dictionary with hyperparameter name as key and hyperparamer value as value.

        """
        default_params = {
            "learning_rate": trial.suggest_float("learning_rate", 0.01, 0.5),
            "alpha": trial.suggest_float("alpha", 0, 1.0),
            "lambda": trial.suggest_float("lambda", 1e-8, 1.0),
            "subsample": trial.suggest_float("subsample", 0.4, 1.0),
            "min_child_weight": trial.suggest_int("min_child_weight", 1, 16),
            "max_depth": trial.suggest_int("max_depth", 3, 10),
            "colsample_bytree": trial.suggest_float("colsample_bytree", 0.5, 1.0),
            "max_delta_step": trial.suggest_int("max_delta_step", 0, 10),
        }

        # Compare the list to the default parameter space
        model_parameters = self.model.get_params()
        keys = [x for x in model_parameters.keys() if x in default_params.keys()]
        # create a dictionary with the matching parameters
        params = {parameter: default_params[parameter] for parameter in keys}

        return params

    def get_pruning_callback(self, trial: optuna.trial.FrozenTrial):
        return None

    def get_trial_track(self) -> dict:
        """Get a dictionary of al trials.

        Returns:
            Dict with al trials and it's parameters

        """
        return self.track_trials

    def create_report(self, model: OpenstfRegressor) -> Report:
        """Generate a report from the data available inside the objective function.

        Args:
            model: OpenstfRegressor, model to create a report on

        Returns:
            Report about the model

        """
        # Report about the training process
        reporter = Reporter(self.train_data, self.validation_data, self.test_data)
        report = reporter.generate_report(model)

        return report

    @classmethod
    def get_default_values(cls) -> dict:
        return {
            "learning_rate": 0.3,
            "alpha": 0.0,
            "lambda": 1.0,
            "subsample": 1.0,
            "min_child_weight": 1,
            "max_depth": 6,
            "colsample_bytree": 1,
            "max_delta_step": 0,
        }


class XGBRegressorObjective(RegressorObjective):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_type = MLModelType.XGB

    # extend the parameters with the model specific ones per implementation
    def get_params(self, trial: optuna.trial.FrozenTrial) -> dict:
        """Get parameters for XGB Regressor Objective with objective specific parameters.

        Args: trial

        Returns:
            Dictionary with hyperparameter name as key and hyperparamer value as value.

        """
        # Filtered default parameters
        model_params = super().get_params(trial)

        # XGB specific parameters
        params = {
            "gamma": trial.suggest_float("gamma", 0.0, 1.0),
            "booster": trial.suggest_categorical("booster", ["gbtree", "dart"]),
        }
        return {**model_params, **params}

    def get_pruning_callback(self, trial: optuna.trial.FrozenTrial):
        return optuna.integration.XGBoostPruningCallback(
            trial, observation_key=f"validation_1-{self.eval_metric}"
        )

    @classmethod
    def get_default_values(cls) -> dict:

        default_parameter_values = super().get_default_values()
        default_parameter_values.update({"gamma": 0.0, "booster": "gbtree"})
        return default_parameter_values


class LGBRegressorObjective(RegressorObjective):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_type = MLModelType.LGB

    def get_params(self, trial: optuna.trial.FrozenTrial) -> dict:
        """Get parameters for LGB Regressor Objective with objective specific parameters.

        Args: trial

        Returns:
            Dictionary with hyperparameter name as key and hyperparamer value as value.

        """
        # Filtered default parameters
        model_params = super().get_params(trial)

        # LGB specific parameters
        params = {
            "num_leaves": trial.suggest_int("num_leaves", 16, 62),
            "boosting_type": trial.suggest_categorical(
                "boosting_type", ["gbdt", "dart", "rf"]
            ),
            "tree_learner": trial.suggest_categorical(
                "tree_learner", ["serial", "feature", "data", "voting"]
            ),
            "n_estimators": trial.suggest_int("n_estimators", 50, 150),
            "min_split_gain": trial.suggest_float("min_split_gain", 1e-8, 1),
            "subsample_freq": trial.suggest_int("subsample_freq", 1, 10),
        }
        return {**model_params, **params}

    def get_pruning_callback(self, trial: optuna.trial.FrozenTrial):
        metric = self.eval_metric
        if metric == "mae":
            metric = "l1"
        return optuna.integration.LightGBMPruningCallback(
            trial, metric=metric, valid_name="valid_1"
        )


class XGBQuantileRegressorObjective(RegressorObjective):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_type = MLModelType.XGB_QUANTILE

    def get_params(self, trial: optuna.trial.FrozenTrial) -> dict:
        """Get parameters for XGBQuantile Regressor Objective with objective specific parameters.

        Args: trial

        Returns:
            Dictionary with hyperparameter name as key and hyperparamer value as value.

        """
        # Filtered default parameters
        model_params = super().get_params(trial)

        # XGB specific parameters
        params = {
            "gamma": trial.suggest_float("gamma", 1e-8, 1.0),
        }
        return {**model_params, **params}

    def get_pruning_callback(self, trial: optuna.trial.FrozenTrial):
        return optuna.integration.XGBoostPruningCallback(
            trial, observation_key=f"validation_1-{self.eval_metric}"
        )


class ProLoafRegressorObjective(RegressorObjective):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_type = MLModelType.ProLoaf

    def get_params(self, trial: optuna.trial.FrozenTrial) -> dict:
        """Get parameters for ProLoaf Regressor Objective with objective specific parameters.

        Args: trial

        Returns:
            Dictionary with hyperparameter name as key and hyperparamer value as value.

        """
        # Filtered default parameters
        model_params = super().get_params(trial)

        # ProLoaf specific parameters
        params = {
            # TODO: look into optimizing this pipeline for proloaf
            # "relu_leak": trial.suggest_float("relu_leak", 0.1, 1.0),
            # "core_layers": trial.suggest_int("core_layers", 1, 3),
            # "rel_linear_hidden_size": trial.suggest_float(
            #    "rel_linear_hidden_size", 0.1, 1
            # ),
            # "rel_core_hidden_size": trial.suggest_float("rel_core_hidden_size", 0.1, 1),
            # "dropout_fc": trial.suggest_float("dropout_fc", 0.1, 0.9),
            # "dropout_core": trial.suggest_float("dropout_core", 0.1, 0.9),
            # "early_stopping_patience": trial.suggest_int(
            #    "early_stopping_patience", 5, 10
            # ),
            # "early_stopping_margin": trial.suggest_float(
            #    "early_stopping_margin", 0.1, 0.9
            # ),
            "max_epochs": trial.suggest_int(
                "max_epochs", 1, 1
            ),  # TODO: change after having availability to gpu resource
            "batch_size": trial.suggest_int("batch_size", 1, 24),
        }
        return {**model_params, **params}

    def get_pruning_callback(self, trial: optuna.trial.FrozenTrial):
        return optuna.integration.PyTorchLightningPruningCallback(
            trial, monitor="val_loss"
        )


class LinearRegressorObjective(RegressorObjective):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_type = MLModelType.LINEAR

    def get_params(self, trial: optuna.trial.FrozenTrial) -> dict:
        """Get parameters for Linear Regressor Objective with objective specific parameters.

        Args: trial

        Returns:
            Dictionary with hyperparameter name as key and hyperparamer value as value.

        """
        # Imputation strategy
        params = {
            "imputation_strategy": trial.suggest_categorical(
                "imputation_strategy", ["mean", "median", "most_frequent"]
            ),
        }
        return params


class ARIMARegressorObjective(RegressorObjective):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_type = MLModelType.ARIMA

    def get_params(self, trial: optuna.trial.FrozenTrial) -> dict:
        """Get parameters for ARIMA Regressor Objective with objective specific parameters.

        Temporary, it seems strange to use optuna for ARIMA models,
        it is usually done via statistical analysis and heuristics.

        Args: trial

        Returns:
            Dictionary with hyperparameter name as key and hyperparamer value as value.

        """
        # Imputation strategy
        params = {
            "trend": trial.suggest_categorical("trend", ["n", "c", "t", "ct"]),
        }
        return params



# SPDX-FileCopyrightText: 2017-2023 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module defines the missing value handler."""
from typing import Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, MetaEstimatorMixin, RegressorMixin, clone
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y


class MissingValuesHandler(BaseEstimator, RegressorMixin, MetaEstimatorMixin):
    """Class for a meta-models that handles missing values and removes columns filled exclusively by NaN.

    It's a pipeline of:

        - An Imputation transformer for completing missing values.
        - A Regressor fitted on the filled data.

    Args:
        base_estimator: Regressor used in the pipeline.
        missing_values: The placeholder for the missing values. All occurrences of
            `missing_values` will be imputed. For pandas' dataframes with
            nullable integer dtypes with missing values, `missing_values`
            should be set to `np.nan`, since `pd.NA` will be converted to `np.nan`.
        imputation_strategy:  The imputation strategy.
            - If None no imputation is performed.
            - If "mean", then replace missing values using the mean along
            each column. Can only be used with numeric data.
            - If "median", then replace missing values using the median along
            each column. Can only be used with numeric data.
            - If "most_frequent", then replace missing using the most frequent
            value along each column. Can be used with strings or numeric data.
            If there is more than one such value, only the smallest is returned.
            - If "constant", then replace missing values with fill_value. Can be
            used with strings or numeric data.
        fill_value: When strategy == "constant", fill_value is used to replace all
            occurrences of missing_values.
            If left to the default, fill_value will be 0 when imputing numerical
            data and "missing_value" for strings or object data types.

    Attributes:
        feature_names:
            All input feature.

        non_null_columns_:
            Valid features used by the regressor.

        n_features_in_:
            Number of input features.

        regressor_: RegressorMixin
            Regressor fitted on valid columns.

        imputer_: SimpleImputer
            Imputer for missig value fitted on valid columns.

        pipeline_: Pipeline
            Pipeline that chains the imputer and the regressor.

        feature_importances_: ndarray (n_features_in_, )
            The feature importances from the regressor for valid features and zero otherwise.

    """

    def __init__(
        self,
        base_estimator: RegressorMixin,
        missing_values: Union[int, float, str, None] = np.nan,
        imputation_strategy: str = None,
        fill_value: Union[str, int, float] = None,
    ):
        """Initialize missing values handler."""
        self.base_estimator = base_estimator
        self.missing_values = missing_values
        self.imputation_strategy = imputation_strategy
        self.fill_value = fill_value

    def _get_tags(self):
        tags = self.base_estimator._get_tags()
        tags["requires_y"] = True
        tags["multioutput"] = False
        tags["allow_nan"] = self.imputation_strategy is not None
        return tags

    def fit(self, x, y):
        """Fit model."""
        _, y = check_X_y(x, y, force_all_finite="allow-nan", y_numeric=True)
        if type(x) != pd.DataFrame:
            x = pd.DataFrame(np.asarray(x))
        self.feature_in_names_ = list(x.columns)
        self.n_features_in_ = x.shape[1]

        # Remove always null columns
        is_column_null = x.isnull().all(axis="index")
        self.non_null_columns_ = list(x.columns[~is_column_null])

        self.regressor_ = clone(self.base_estimator)

        # Build the proper imputation transformer
        # - Identity function if strategy is None
        # - SimpleImputer with the dedicated strategy
        if self.imputation_strategy is None:
            self.imputer_ = FunctionTransformer(func=self._identity)
        else:
            self.imputer_ = SimpleImputer(
                missing_values=self.missing_values,
                strategy=self.imputation_strategy,
                fill_value=self.fill_value,
            )

        self.pipeline_ = Pipeline(
            [("imputer", self.imputer_), ("regressor", self.regressor_)]
        )

        # Fit only on non_null_columns
        self.pipeline_.fit(x[self.non_null_columns_], y)

        return self

    @classmethod
    def _identity(cls, x):
        return x

    def predict(self, x):
        """Make a prediction."""
        check_is_fitted(self)
        check_array(
            x,
            force_all_finite="allow-nan",
        )
        if type(x) != pd.DataFrame:
            x = pd.DataFrame(np.array(x))
        return self.pipeline_.predict(x[self.non_null_columns_])



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import numpy as np
import pandas as pd
from sklearn.base import RegressorMixin


class StandardDeviationGenerator:
    def __init__(self, validation_data: pd.DataFrame) -> None:

        self.validation_data = validation_data

    def generate_standard_deviation_data(self, model: RegressorMixin) -> RegressorMixin:
        """Generate the standard data.

        Calculates the difference between realised and predicted on validation set.
        For each hour of the day the std of the difference is calculated.

        Args:
            model: The trained model

        Returns:
            The model with the std data added.

        """
        # Define some variables
        predicted = None
        self.standard_deviation = pd.DataFrame()

        # Loop over horizons and ask prediction for each specific horizon
        for horizon in self.validation_data.horizon.unique():
            # Make subset for this specific horizon
            sub_val = self.validation_data[self.validation_data.horizon == horizon]
            try:
                predicted = model.predict(sub_val.iloc[:, 1:-1])
            except Exception as e:
                print("Could not get prediction from new model!", e)

            # Calculate confidence interval for this horizon
            confidence_interval_horizon = self._calculate_standard_deviation(
                sub_val.iloc[:, 0], predicted
            )
            confidence_interval_horizon[
                "horizon"
            ] = horizon  # Label with respective horizon
            self.standard_deviation = pd.concat(
                [self.standard_deviation, confidence_interval_horizon]
            )

        model.standard_deviation = self.standard_deviation

        return model

    @staticmethod
    def _calculate_standard_deviation(
        realised: pd.Series, predicted: pd.Series
    ) -> pd.DataFrame:
        """Protected static method to calculate the corrections for a model.

        Args:
            realised: pd.series with realised load
            predicted: pd.series with load predicted by new model

        Returns:
            DataFrame with model corrections

        """
        result = pd.DataFrame(index=range(24), columns=["stdev", "hour"])
        # Calculate the error for each predicted point
        error = realised - predicted
        error.index = error.index.hour  # Hour only, remove the rest
        # For the time starts with 00, 01, 02, etc. TODO (MAKE MORE ELEGANT SOLUTION THAN A LOOP)
        for hour in range(24):
            hour_error = error[error.index == hour]
            result["stdev"].iloc[hour] = np.std(
                hour_error[1:]
            )  # Exclude first item as this is the hour itself!
            result["hour"].iloc[hour] = hour

        result = result.astype("float")

        return result



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains all lag features."""
import re

import numpy as np
import pandas as pd
import scipy.signal


def generate_lag_feature_functions(
    feature_names: list[str] = None, horizon: float = 24.0
) -> dict:
    """Creates functions to generate lag features in a dataset.

    Args:
        feature_names: minute lagtimes that where used during training
            of the model. If empty a new set will be automatically generated.
        horizon: Forecast horizon limit in hours.

    Returns:
        Lag functions.

    Example:

    .. code-block:: py

        lag_functions = generate_lag_functions(data,minute_list,h_ahead)

    """
    # Use extracted lag features if provided.
    if feature_names is not None:
        lag_times_minutes, lag_time_days_list = extract_lag_features(
            feature_names, horizon
        )
    else:
        # Generate available lag_times if no features are provided
        lag_times_minutes, lag_time_days_list = generate_trivial_lag_features(horizon)

    # Empty dict to store all generated lag functions
    lag_functions = {}
    for minutes in lag_times_minutes:  # Add intraday-lag functions (lags in minutes)

        def func(x, shift=minutes):
            return x.shift(freq="T", periods=1 * shift)

        new = {"T-" + str(int(minutes)) + "min": func}
        lag_functions.update(new)

    # Add day lag functions:
    for day in lag_time_days_list:

        def func(x, shift=day):
            return x.shift(freq="1d", periods=1 * shift)

        new = {"T-" + str(int(day)) + "d": func}
        lag_functions.update(new)
    return lag_functions


def extract_lag_features(
    feature_names: list[str], horizon: float = 24.0
) -> tuple[list, list]:
    """Creates a list of lag minutes and a list of lag days that were used during the training of the input model.

    Args:
        feature_names: All requested lag features
        horizon: Forecast horizon limit in hours.

    Returns:
        - List of minute lags that were used as features during training.
        - List of days lags that were used as features during training.

    """
    # Prepare empty lists to append on
    minutes_list = []
    days_list = []

    for lag_feature in feature_names:

        # Select the number of days or the number of minutes by matching with a regular expression
        number_of_minutes = re.search(r"T-(\d+)min", lag_feature)
        number_of_days = re.search(r"T-(\d+)d", lag_feature)

        # Append to the appropriate list
        if number_of_minutes is not None:
            minutes_list.append(int(number_of_minutes[1]))
        elif number_of_days is not None:
            days_list.append(int(number_of_days[1]))

    # Discard lag times that are not available for the specified horizon
    minutes_list = list(set([i for i in minutes_list if i >= horizon * 60]))
    days_list = list(set([i for i in days_list if i >= horizon / 24]))

    return minutes_list, days_list


def generate_trivial_lag_features(horizon: float) -> tuple[list, list]:
    """Generates relevant lag times for lag feature function creation.

    This function is mostly used during training of models and not during predicting.

    Args:
        horizon: Forecast horizon limit in hours.

    Returns:
        - List of minute lags that were used as features during training.
        - List of days lags that were used as features during training.

    """
    mindays = int(np.ceil(horizon / 24))
    lag_time_days_list = list(np.linspace(mindays, 14, 15 - mindays))

    # Make list of trivial lag times
    trivial_lag_minutes_list = np.linspace(60, 23 * 60, 23).tolist() + [15, 30, 45]

    # Discard lag times that are not available for the specified horizon
    trivial_lag_times_minutes = list(
        set([i for i in trivial_lag_minutes_list if i >= horizon * 60])
    )

    return trivial_lag_times_minutes, lag_time_days_list


def generate_non_trivial_lag_times(
    data: pd.DataFrame, height_treshold: float = 0.1
) -> list[int]:
    """Calculate an autocorrelation curve of the load trace.

    This curve is subsequently used to add additional lag times as features.

    Args:
        data: Dataframe with input data in the form pd.DataFrame(index = datetime,
                             columns = [label, predictor_1,..., predictor_n])
        height_treshold: Minimal autocorrelation value to be recognized as a peak.

    Returns:
        Aditional non-trivial minute lags

    """

    def autocorr(x: np.array, lags: range) -> np.array:
        """Make an autocorrelation curve."""
        mean = x.mean()
        var = np.var(x)
        xp = x - mean
        corr = np.correlate(xp, xp, "full")[len(x) - 1 :] / var / len(x)

        return corr[: len(lags)]

    try:
        # Get rid of nans as the autocorrelation handles these values badly
        data = data[data.columns[0]].dropna()  # First column contains the load
        # Get autocorrelation curve
        y = autocorr(data, range(10000))
        # Determine the peaks (positive and negative) larger than a specified threshold
        peaks = scipy.signal.find_peaks(np.abs(y), height=height_treshold)
        peaks = peaks[0]
        # Convert peaks to lag times in minutes
        peaks = peaks[peaks < (60 * 4)]
        additional_minute_space = peaks * 15
    except Exception:
        return []
    # Return list of additional minute lags to be procceses by apply features
    return list(additional_minute_space)
